#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import shutil
import argparse
from pathlib import Path

import numpy as np
import networkx as nx

from tqdm import tqdm

from ase import Atoms
from ase.io import read, write


"""This temporarily stores some utilities for atoms object.
"""

def check_convergence(atoms, fmax=0.05):
    """Check the convergence of the trajectory"""

    forces = atoms.get_forces()

    max_force = np.max(np.fabs(forces))

    converged = False
    if max_force < fmax:
        converged = True 

    return converged

def merge_xyz():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p", "--pattern",
        help="pattern to find xyz files"
    )
    args = parser.parse_args()

    #pattern = "s2hcp"
    pattern = args.pattern
    cwd = Path.cwd()

    sorted_paths = []
    for p in cwd.glob(pattern+"*.xyz"):
        sorted_paths.append(p)
    sorted_paths.sort()

    frames = []
    for p in sorted_paths:
        cur_frames = read(p, ":")
        print(p.stem)
        for atoms in cur_frames:
            atoms.info["source"] = str(p.stem)
        print(p, " #frames: ", len(cur_frames))
        frames.extend(cur_frames)
    print("TOTAL #frames: ", len(frames))
    write(pattern+"-0906.xyz", frames)

    for p in sorted_paths:
        shutil.move(p, cwd / ("bak."+p.name))


def sort_atoms(atoms):
    # sort atoms by symbols and z-positions especially for supercells 
    numbers = atoms.numbers 
    zposes = atoms.positions[:,2].tolist()
    sorted_indices = np.lexsort((zposes,numbers))
    atoms = atoms[sorted_indices]

    return atoms


def try_sort():
    frames = read('./test_data.xyz', ':')
    
    new_frames = [sort_atoms(atoms) for atoms in frames]
    write('new_test.xyz', new_frames)


def get_structure_type(atoms_: Atoms) -> str:
    """Determine the structure type based on atom connectivity.

    For bulk, there are bond connections in xyz 3 dimensions. For surface, there 
    are bond connections in one dimensions (default should be z-axis). For cluster
    or molecue, there are no bond connections among neighbour cells.

    Returns:
        bulk, surface, and cluster

    """
    #natoms = len(atoms)

    #from ase.neighborlist import natural_cutoffs, NeighborList
    #cutoffs = natural_cutoffs(atoms, mult=1.0)
    #neigh = NeighborList(
    #    cutoffs, skin=0.0, sorted=False, self_interaction=True, bothways=False,
    #)
    #neigh.update(atoms)

    #stype = "cluster"
    #has_x, has_y, has_z = False, False, False
    #for i in range(natoms):
    #    indices, offsets = neigh.get_neighbors(i)
    #    for j, offset in zip(indices, offsets):
    #        if offset[0] != 0:
    #            has_x = True
    #        if offset[1] != 0:
    #            has_y = True
    #        if offset[2] != 0:
    #            has_z = True
    #        print(f"{i}-{j}: ", offset, [has_x, has_y, has_z])
    #    if has_x and has_y and has_z:
    #        stype = "bulk"
    #        break
    #if sum([has_x, has_y, has_z]) == 2:
    #    stype = "surface"
    #    # TODO: find in which direction is the vaccum
    #    ...
    #print([has_x, has_y, has_z])
    atoms = copy.deepcopy(atoms_)
    #stype = atoms.info.get("stype")

    # -
    graph = create_multipgraph(atoms)
    # NOTE: A cluster has atoms concentrated around the centre.
    is_cluster = True
    # TODO: water box?
    #print(graph)
    for u, v, d in graph.edges.data():
        #print(u, v, d)
        # NOTE: at least a surface if there are bonds between neighbour boxs
        # TODO: a single molecule crosses the boundary?
        if sum(d["offset"]) != 0:
            is_cluster = False
            break

    if not is_cluster:
        # --- check if surface or bulk
        graph = create_multipgraph(atoms, pbc=False)
        subgraphs = [graph.subgraph(nodes) for nodes in nx.connected_components(graph)]
        #print(subgraphs)
        # TODO: water box?
        cell_ = atoms.get_cell(complete=True)
        if len(subgraphs) == 1:
            #stype = "bulk"
            positions_ = atoms.get_positions()
        else:
            cops = [
                np.average(atoms.positions[g.nodes], axis=0).tolist()
                for g in subgraphs
            ]
            orders = [x[0] for x in sorted(enumerate(cops), key=lambda x: x[1][2])]
            #print(cops)
            #print(orders)
            positions_ = copy.deepcopy(atoms.get_positions())
            for i in orders[1:]:
                for a_idx in subgraphs[i].nodes:
                    positions_[a_idx] += np.dot(cell_, [0,0,-1])
            # NOTE: avoid some bulk systems that they are moved far below
            if np.average(positions_, axis=0)[2] < 0.:
                positions_ = copy.deepcopy(atoms.get_positions())

        zvec = np.cross(cell_[0], cell_[1])
        zlength = np.dot(zvec/np.linalg.norm(zvec), cell_[2])
        max_zpos = np.max(positions_[:,2])
        #print("max_zpos: ", max_zpos)
        if max_zpos + 6.0 <= zlength: # NOTE: larger than a cutoff
            stype = "surface"
        else:
            stype = "bulk"
        #print(stype)
    else:
        stype = "cluster"

    return stype

def create_multipgraph(atoms_, pbc=True):
    """"""
    atoms = copy.deepcopy(atoms_)

    graph = nx.MultiGraph(atoms=atoms)
    #print(graph.graph)

    # - find neighbours
    natoms = len(atoms)
    atoms.pbc = pbc
    #atoms.pbc = False
    #atoms.pbc = [True, True, False]

    from ase.neighborlist import natural_cutoffs, NeighborList
    cutoffs = natural_cutoffs(atoms, mult=1.2)
    #cutoffs = [6.0]*natoms
    neigh = NeighborList(
        cutoffs, skin=0.0, sorted=False, self_interaction=False, bothways=False,
    )
    neigh.update(atoms)

    # -- add nodes
    nodes = [[i, {"symbol": a.symbol}] for i, a in enumerate(atoms)]
    graph.add_nodes_from(nodes)
    #print(graph)
    #for u, d in graph.nodes.data():
    #    print(u, d)

    # -- add edges
    edges = []
    for i in range(natoms):
        indices, offsets = neigh.get_neighbors(i)
        for j, offset in zip(indices, offsets):
            edges.append([i, j, {"route": 2, "offset": offset.tolist()}])
    graph.add_edges_from(edges)

    #print("--- edges ---") # find egdes without offset [0,0,0]
    #for u, v, d in graph.edges.data():
    #    print(u, v, d)
    #for u, v, keys in graph.edges(keys=True):
    #    print(u, v, keys)
    #for n, nbrsdict in graph.adjacency():
    #    print(n, nbrsdict)
    #    for nbr, keydict in nbrsdict.items():
    #        print(nbr, keydict)
    #        for key, eattr in keydict.items():
    #            ...
    #print(graph.edges())
    #print(set(graph.edges()))
    #for u, v, d in graph.edges.data(nbunch=[10]):
    #    print(u, v, d)
    
    # -- find subgraphs
    #subgraphs = nx.connected_components(graph)
    #print(subgraphs)

    return graph


if __name__ == "__main__":
    """"""
    stru_dict = {}

    frames = read("/home/jx1279/projects/sintering/dataset/Cu12.xyz", ":")
    #frames = read("/home/jx1279/projects/sintering/dataset/Cu16.xyz", ":")
    #frames = read("/home/jx1279/projects/sintering/dataset/Cu87.xyz", ":")
    nframes = len(frames)
    print("nframes: ", nframes)

    for i, atoms in tqdm(enumerate(frames)):
        atoms.info["confid"] = i
        # NOTE: cluster/molecule (0d), nanowire (1d), surface (2d), bulk (3d)
        # TODO: determine by volume?
        stype = get_structure_type(atoms)

        # -- 
        name = f"{atoms.get_chemical_formula()}_{stype}"
        if name in stru_dict:
            stru_dict[name].append(atoms)
        else:
            stru_dict[name] = [atoms]
    
    new_nframes = 0
    for (name, cur_frames) in stru_dict.items():
        new_nframes += len(cur_frames)
        print(len(cur_frames))
        write(f"./{name}.xyz", cur_frames)
    print(new_nframes)
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import traceback

from .. import config
from ..core.register import registers
from ..builder.builder import StructureBuilder
from ..builder.utils import convert_string_to_atoms
from ..worker.single import SingleWorker
from ..worker.drive import DriverBasedWorker
from ..worker.interface import ComputerVariable
from ..utils.command import convert_indices, dict2str

from .interface import ExpeditionVariable
registers.variable.register(ExpeditionVariable)

from .interface import explore 
registers.operation.register(explore)

from .expedition import AbstractExpedition

from .ga.engine import GeneticAlgorithemEngine
registers.expedition.register("genetic_algorithm")(GeneticAlgorithemEngine)

from .monte_carlo.basin_hopping import BasinHopping
registers.expedition.register("basin_hopping")(BasinHopping)

from .monte_carlo.monte_carlo import MonteCarlo
registers.expedition.register("monte_carlo")(MonteCarlo)

from .simulated_annealing.simulated_annealing import SimulatedAnnealing
registers.expedition.register("simulated_annealing")(SimulatedAnnealing)

# - optional
try:
    from .af.afir import AFIRSearch
    registers.expedition.register("artificial_reaction")(AFIRSearch)
except:
    config._print("AFIR is not loaded.")
    config._print(traceback.print_exc())


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


try:
    import jax
except Exception as e:
    ...

from ..core.register import registers

from .distance import DistanceColvar
registers.colvar.register("DistanceColvar")(DistanceColvar)

from .fingerprint import FingerprintColvar
registers.colvar.register("FingerprintColvar")(FingerprintColvar)

from .position import position
registers.colvar.register("position")(position)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from ..core.register import registers

"""Add bias on potential energy surface. 

Some of bias forces are based on JAX. In the future, we need replace those oft-used
ones to pure python codes as jax need accelerate them a lot.

"""

from .afir import AFIRCalculator
registers.bias.register("afir")(AFIRCalculator)

from .bondboost import BondBoostCalculator
registers.bias.register("bondboost")(BondBoostCalculator)

# from .harmonic import HarmonicBias
# bias_register.register("harmonic")(HarmonicBias)

from .harmonic import DistanceHarmonicCalculator, PlaneHarmonicCalculator
registers.bias.register("distance_harmonic")(DistanceHarmonicCalculator)
registers.bias.register("plane_harmonic")(PlaneHarmonicCalculator)

# - gaussian

# from .gaussian import GaussianCalculator
# bias_register.register("gaussian")(GaussianCalculator)

from .gaussian import BondGaussianCalculator, CenterOfMassGaussianCalculator, RMSDGaussian
registers.bias.register("bond_gaussian")(BondGaussianCalculator)
registers.bias.register("center_of_mass_gaussian")(CenterOfMassGaussianCalculator)
registers.bias.register("rmsd_gaussian")(RMSDGaussian)


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.geometry import find_mic

import ase.optimize
from ase.constraints import FixAtoms
from ase.constraints import UnitCellFilter

from ase.calculators.singlepoint import SinglePointCalculator

from gdpx.validator.validator import AbstractValidator
from gdpx.worker.drive import DriverBasedWorker
from gdpx.builder.constraints import set_constraint
from ..data.array import AtomsNDArray

"""Validate minima and relative energies...
"""

def make_clean_atoms(atoms_, results=None):
    """Create a clean atoms from the input."""
    atoms = Atoms(
        symbols=atoms_.get_chemical_symbols(),
        positions=atoms_.get_positions().copy(),
        cell=atoms_.get_cell().copy(),
        pbc=copy.deepcopy(atoms_.get_pbc())
    )
    if results is not None:
        spc = SinglePointCalculator(atoms, **results)
        atoms.calc = spc

    return atoms


class MinimaValidator(AbstractValidator):

    """Run minimisation on various configurations and compare relative energy.

    TODO: 

        Support the comparison of minimisation trajectories.

    """

    def __init__(self, ene_shift=[], *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        self.ene_shift = ene_shift

        return
    
    def _process_data(self, data) -> List[Atoms]:
        """"""
        data = AtomsNDArray(data)

        # We need a List of Atoms (ndim=1).
        if data.ndim == 1:
            data = data.tolist()
        #elif data.ndim == 2: # assume it is from extract_cache...
        #    data = data.tolist()
        #elif data.ndim == 3: # assume it is from a compute node...
        #    data_ = []
        #    for d in data[:]: # TODO: add squeeze method?
        #        data_.extend(d)
        #    data = data
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data

    def run(self, dataset: dict, worker: DriverBasedWorker, *args, **kwargs):
        """"""
        # TODO: assume dataset is a dict of frames
        pre_dataset = {}
        for k, curr_data in dataset.items():
            curr_frames = self._process_data(curr_data)
            nframes = len(curr_frames)

            cached_pred_fpath = self.directory/k/ "pred.xyz"
            if not cached_pred_fpath.exists():
                worker.directory = self.directory/k
                worker.batchsize = nframes

                worker.run(curr_frames)
                worker.inspect(resubmit=True)
                if worker.get_number_of_running_jobs() == 0:
                    # -- get end frames
                    trajectories = worker.retrieve(
                        include_retrieved=True,
                    )
                    pred_frames = [t[-1] for t in trajectories]
                else:
                    # TODO: ...
                    ...
                write(cached_pred_fpath, pred_frames)
            else:
                pred_frames = read(cached_pred_fpath, ":")
            
            pre_dataset[k] = pred_frames
        
        # -
        keys = list(dataset.keys())
        for key in keys:
            self._print(f"group {key}")
            # - restore constraints
            cons_text = worker.driver.as_dict()["constraint"]
            for ref_atoms, pre_atoms in zip(dataset[key], pre_dataset[key]):
                set_constraint(ref_atoms, cons_text, ignore_attached_constraints=True)
                set_constraint(pre_atoms, cons_text, ignore_attached_constraints=True)
        
            # - compare ...
            ref_energies = np.array([a.get_potential_energy() for a in dataset[key]])
            pre_energies = np.array([a.get_potential_energy() for a in pre_dataset[key]])

            ref_maxforces = np.array([np.max(np.fabs(a.get_forces(apply_constraint=True))) for a in dataset[key]])
            pre_maxforces = np.array([np.max(np.fabs(a.get_forces(apply_constraint=True))) for a in pre_dataset[key]])

            # - compute shifts if any
            if key == "composites":
                if self.ene_shift:
                    ref_ene_shift = np.sum([x*dataset[y][0].get_potential_energy() for x, y in self.ene_shift])
                    pre_ene_shift = np.sum([x*pre_dataset[y][0].get_potential_energy() for x, y in self.ene_shift])
            else:
                ref_ene_shift = 0.
                pre_ene_shift = 0.

            ref_rel_energies = ref_energies - ref_ene_shift
            pre_rel_energies = pre_energies - pre_ene_shift

            # -- disp
            disps = [] # displacements
            for ref_atoms, pre_atoms in zip(dataset[key], pre_dataset[key]):
                vector = pre_atoms.get_positions() - ref_atoms.get_positions()
                vmin, vlen = find_mic(vector, pre_atoms.get_cell())
                disps.append(np.linalg.norm(vlen))

            content = ("{:<24s}  "+"{:<12s}  "*7+"\n").format("Name", "RefEne", "NewEne", "RefMaxF", "NewMaxF", "Drmse", "RefRelEne", "NewRelEne")
            for i, (ref_ene, new_ene, ref_maxfrc, new_maxfrc, dis, ref_rel_ene, pre_rel_ene) in enumerate(
                zip(ref_energies, pre_energies, ref_maxforces, pre_maxforces, disps, ref_rel_energies, pre_rel_energies)
            ):
                content += ("{:<24s}  "+"{:<12.4f}  "*7+"\n").format(
                    str(i), ref_ene, new_ene, ref_maxfrc, new_maxfrc, dis, ref_rel_ene, pre_rel_ene
                )

            with open(self.directory/f"abs-{key}.dat", "w") as fopen:
                fopen.write(content)
        
            self._print(content)

        return


if __name__ == "__main__":
    ...


""" Custom Nose-Hoover NVT thermostat based on ASE.

This code was originally written by Jonathan Mailoa based on these notes:

    https://www2.ph.ed.ac.uk/~dmarendu/MVP/MVP03.pdf

It was then adapted by Simon Batzner to be used within ASE. Parts of the overall outline of the class are also based on the Langevin class in ASE.

This was further changed by Jiayan XU for a more compact formulation.

"""

import numpy as np

from ase.md.md import MolecularDynamics
from ase.md.velocitydistribution import Stationary
from ase import units


class NoseHoover(MolecularDynamics):
    """Nose-Hoover (constant N, V, T) molecular dynamics.

    Usage: NoseHoover(atoms, dt, temperature)

    atoms
        The list of atoms.

    timestep
        The time step.

    temperature
        Target temperature of the MD run in [K * units.kB]

    nvt_q
        Q in the Nose-Hoover equations

    Example Usage:

        nvt_dyn = NoseHoover(
            atoms=atoms,
            timestep=0.5 * units.fs,
            temperature=300. * units.kB,
            nvt_q=334.
        )

    """

    def __init__(
        self,
        atoms,
        timestep,
        temperature,
        nvt_q,
        trajectory=None,
        logfile=None,
        loginterval=1,
        append_trajectory=False,
    ):
        # set com momentum to zero
        # TODO: if make com fixed at each step?
        # Stationary(atoms)

        self.temp = temperature / units.kB
        self.nvt_q = nvt_q
        self.dt = timestep  # units: A/sqrt(u/eV)
        self.dtdt = np.power(self.dt, 2)
        self.nvt_bath = 0.0

        # local
        self._vel_halfstep = None

        MolecularDynamics.__init__(
            self,
            atoms,
            timestep,
            trajectory,
            logfile,
            loginterval,
            append_trajectory=append_trajectory,
        )

    def step(self, f=None):
        """ Perform a MD step. 

        """

        # TODO: we do need the f=None argument?
        atoms = self.atoms
        masses = atoms.get_masses()  # units: u

        # count actual degree of freedoms
        # count every step because sometimes constraints can be manually changed
        ndof = 3*len(atoms)
        for constraint in atoms._constraints:
            ndof -= constraint.get_removed_dof(atoms)
        ndof += 1 # bath

        if f is None:
            f = atoms.get_forces()

        # for the first step, v(t-dt/2) = v(0.5) is needed.
        if self._vel_halfstep is None:
            self._vel_halfstep = (
                atoms.get_velocities() - 
                0.5 * self.dt * (f / masses[:, np.newaxis])
            )
        else:
            pass

        # v(t-dt/2), f(t), eta(t) -> v(t)
        atoms.set_velocities(
            (
                self._vel_halfstep
                + 0.5 * self.dt * (f / masses[:, np.newaxis])
            )
            / (1 + 0.5 * self.dt * self.nvt_bath)
        )

        # v(t), f(t), eta(t) -> r(t+dt)
        modified_acc = (
            f / masses[:, np.newaxis]
            - self.nvt_bath * atoms.get_velocities()
        )

        pos_fullstep = (
            atoms.get_positions()
            + self.dt * atoms.get_velocities()
            + 0.5 * self.dtdt * modified_acc
        )

        atoms.set_positions(pos_fullstep)

        # v(t), f(t), eta(t) -> v(t+dt/2)
        self._vel_halfstep = atoms.get_velocities() + 0.5 * self.dt * modified_acc

        # eta(t), v(t) -> eta(t+dt/2)
        e_kin_diff = 0.5 * (
            np.sum(masses * np.sum(atoms.get_velocities() ** 2, axis=1))
            - (ndof) * units.kB * self.temp
        ) # number of freedoms?

        nvt_bath_halfstep = self.nvt_bath + 0.5 * self.dt * e_kin_diff / self.nvt_q

        # eta(t+dt/2), v(t+dt/2) -> eta(t+dt)
        e_kin_diff_halfstep = 0.5 * (
            np.sum(masses * np.sum(self._vel_halfstep ** 2, axis=1))
            - (ndof) * units.kB * self.temp
        )

        self.nvt_bath = (
            nvt_bath_halfstep + 0.5 * self.dt * e_kin_diff_halfstep / self.nvt_q
        )

        return

if __name__ == '__main__':
    import ase.io
    import ase.units
    from ase import Atoms
    from ase.calculators.emt import EMT
    from ase.md.velocitydistribution import MaxwellBoltzmannDistribution

    from ase.constraints import FixAtoms

    from .md_utils import force_temperature

    # ===== system =====
    atoms = Atoms(
        'CO2',
        positions = [
            [0.,0.,0.],
            [0.,0.,1.0],
            [0.,0.,-1.0]
        ]
    )

    cons = FixAtoms(mask=[atom.symbol == 'C' for atom in atoms])

    atoms.set_constraint(cons)

    atoms.set_calculator(EMT())

    # ===== molecular dynamics =====
    temperature = 300
    MaxwellBoltzmannDistribution(atoms, temperature*ase.units.kB)
    force_temperature(atoms, temperature)
    print('start!!!')
    print(atoms.get_velocities())
    print(atoms.get_temperature())

    nvt_dyn = NoseHoover(
        atoms = atoms,
        timestep = 2.0 * units.fs,
        temperature = temperature * units.kB,
        nvt_q = 334.
    )

    def print_temperature(atoms):
        content = 'temperature %8.4f' %atoms.get_temperature()
        print(content)

    nvt_dyn.attach(print_temperature, atoms=atoms)

    nvt_dyn.run(steps=5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings


from ..core.register import registers

from .basin import BasinSelector

registers.selector.register(BasinSelector)

from .compare import CompareSelector

registers.selector.register(CompareSelector)

from .interval import IntervalSelector

registers.selector.register(IntervalSelector)

from .invariant import InvariantSelector

registers.selector.register(InvariantSelector)

from .locate import LocateSelector

registers.selector.register(LocateSelector)

from .property import PropertySelector

registers.selector.register(PropertySelector)

from .random import RandomSelector

registers.selector.register(RandomSelector)

from .scf import ScfSelector

registers.selector.register(ScfSelector)

try:
    # TODO: This selector depends on an external package dscribe.
    from .descriptor import DescriptorSelector

    registers.selector.register(DescriptorSelector)
except ImportError as e:
    warnings.warn(f"Module DescriptorSelector import failed: {e}", UserWarning)


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import pickle 

import numpy as np

from ase.io import read, write 

import matplotlib as mpl
mpl.use('Agg') #silent mode
from matplotlib import pyplot as plt


def rms_dict(x_ref, x_pred):
    """ Takes two datasets of the same shape and returns a dictionary containing RMS error data"""

    x_ref = np.array(x_ref)
    x_pred = np.array(x_pred)

    if np.shape(x_pred) != np.shape(x_ref):
        raise ValueError('WARNING: not matching shapes in rms')

    error_2 = (x_ref - x_pred) ** 2

    average = np.sqrt(np.average(error_2))
    std_ = np.sqrt(np.var(error_2))

    return {'rmse': average, 'std': std_}

def energy_plot(ener_in, ener_out, ax, title='Plot of energy'):
    """ Plots the distribution of energy per atom on the output vs the input"""
    # scatter plot of the data
    ax.scatter(ener_in, ener_out)

    # get the appropriate limits for the plot
    for_limits = np.array(ener_in +ener_out)
    elim = (for_limits.min() - 0.05, for_limits.max() + 0.05)
    ax.set_xlim(elim)
    ax.set_ylim(elim)

    # add line of slope 1 for refrence
    ax.plot(elim, elim, c='k')

    # set labels
    ax.set_ylabel('energy by GAP / eV')
    ax.set_xlabel('energy by VASP / eV')

    #set title
    ax.set_title(title)

    # add text about RMSE
    _rms = rms_dict(ener_in, ener_out)
    rmse_text = 'RMSE:\n' + str(np.round(_rms['rmse'], 3)) + ' +- ' + str(np.round(_rms['std'], 3)) + 'eV/atom'
    ax.text(0.9, 0.1, rmse_text, transform=ax.transAxes, fontsize='large', \
            horizontalalignment='right', verticalalignment='bottom')

def force_plot(in_force, out_force, ax, symbol='H', title='Plot of force'):
    """ Plots the distribution of force components per atom 
        on the output vs the input only plots for the given atom type(s)"""
    # extract data for only one species
    in_force = in_force[symbol]
    out_force = out_force[symbol]

    # scatter plot of the data
    ax.scatter(in_force, out_force)

    # get the appropriate limits for the plot
    for_limits = np.array(in_force + out_force)
    flim = (for_limits.min() - 1, for_limits.max() + 1)
    ax.set_xlim(flim)
    ax.set_ylim(flim)

    # add line of
    ax.plot(flim, flim, c='k')

    # set labels
    ax.set_ylabel('force by GAP / (eV/Å)')
    ax.set_xlabel('force by VASP / (eV/Å)')

    #set title
    ax.set_title(title)

    # add text about RMSE
    _rms = rms_dict(in_force, out_force)
    rmse_text = 'RMSE:\n' + str(np.round(_rms['rmse'], 3)) + ' +- ' + str(np.round(_rms['std'], 3)) + 'eV/Å'
    ax.text(0.9, 0.1, rmse_text, transform=ax.transAxes, fontsize='large', horizontalalignment='right',
        verticalalignment='bottom')


def extract_energy_and_forces(atom_frames,calc=None,atomic=True):
    """
    Electronic free energy and Hellman-Feynman forces
    """
    energies_dft, forces_dft = [], {}
    energies_gap, forces_gap = [], {}

    for atoms in atom_frames: # free energy per atom
        # basic info
        symbols = atoms.get_chemical_symbols()
        if atomic:
            natoms = len(atoms)
        else:
            natoms = 1
        # energy
        free_energy = atoms.get_potential_energy(force_consistent=True) # electronic free energy
        energies_dft.append(free_energy/natoms)
        # force
        forces = atoms.get_forces()
        for sym, force in zip(symbols,forces):
            if sym in forces_dft.keys():
                forces_dft[sym].extend(force)
            else:
                forces_dft[sym] = list(force)
        if calc:
            # use quip to calculate gap predicted energy
            atoms.set_calculator(calc)
            free_energy = atoms.get_potential_energy(force_consistent=True) # electronic free energy
            energies_gap.append(free_energy/natoms)
            # force
            forces = atoms.get_forces()
            for sym, force in zip(symbols,forces):
                if sym in forces_gap.keys():
                    forces_gap[sym].extend(force)
                else:
                    forces_gap[sym] = list(force)

    return forces_dft, forces_gap, energies_dft, energies_gap


if __name__ == '__main__':
    # parser
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-t', '--train', 
        default='evaluated.xyz', help='trained structures in xyz format'
    )
    parser.add_argument(
        '-c', '--calc', action='store_true', 
        help='if using dp'
    )
    parser.add_argument(
        '-g', '--graph', 
        default='graph.pb', help='deep potential'
    )

    args = parser.parse_args()

    # calculate using dp 
    if args.calc:
        frames = read(args.train, ':')

        from deepmd.calculator import DP 
        calc = DP(
            model = args.graph, 
            #type_dict = {"O": 1, "Zn": 2, "Cr": 0}
        )
        data_dict = {'ener': [[],[]], 'force': [[],[]]}
        for atoms in frames:
            natoms = len(atoms) 
            energy = atoms.get_potential_energy() / natoms
            forces = atoms.get_forces() 
            data_dict['ener'][0].append(energy)
            data_dict['force'][0].extend(forces.flatten().tolist())
            # calculation 
            calc.reset() 
            atoms.calc = calc 
            energy = atoms.get_potential_energy() / natoms
            forces = atoms.get_forces()
            data_dict['ener'][1].append(energy)
            data_dict['force'][1].extend(forces.flatten().tolist())
            #print(dp_data)
        #print(tot_dat)

        with open('prop.pkl', 'wb') as fopen:
            pickle.dump(data_dict, fopen)

        print('finish calculating...')
    else:
        with open('prop.pkl', 'rb') as fopen:
            tot_data = pickle.load(fopen)
        ener_dat = tot_data['ener']
        force_dat = tot_data['force']
        force_dat[0] = {'Cu': force_dat[0]}
        force_dat[1] = {'Cu': force_dat[1]}
    
        # plot
        fig, axarr = plt.subplots(
            nrows=1, ncols=2, 
            gridspec_kw={'hspace': 0.3}, figsize=(12,8)
        )
        axarr = axarr.flat[:]

        plt.suptitle('Electronic Free Energy and Hellman-Feynman Forces')

        # plot energy
        energy_plot(
            ener_dat[0], ener_dat[1], 
            axarr[0], 'Energy on training data'
        )
        force_plot(
            force_dat[0], force_dat[1], 
            axarr[1], 'Cu', 'Force on training data - Cu'
        )

        plt.savefig('benchmark.png')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import time

from pathlib import Path
from collections import namedtuple

import numpy as np

from ase.io import read, write

from tqdm import tqdm

from joblib import Parallel, delayed

def extract_mctraj():
    suspects = Path("./suspects")
    step_dirs = []
    #step_dirs = list(suspects.glob("step[0-9]"))
    #step_dirs.extend(list(suspects.glob("step[0-9][0-9]")))
    #step_dirs.extend(list(suspects.glob("step[0-9][0-9][0-9]")))
    #step_dirs.extend(list(suspects.glob("step[0-9][0-9][0-9][0-9]")))
    
    for p in suspects.iterdir():
        if re.match(r"step[0-9]{1,4}$", str(p.name)):
            step_dirs.append(p)
    print(len(step_dirs))
    
    step_dirs_sorted = sorted(step_dirs, key=lambda k: int(k.name[4:])) # sort by name
    #print(step_dirs_sorted)
    
    StepInfo = namedtuple("StepInfo", ["natoms", "energy", "deviation"])
    
    def extract_data(p):
        # total energy deviation
        devi_file = p / "model_devi.out"
        devi_info = np.loadtxt(devi_file) # EANN
        en_devi = float(devi_info[1])
        #print(en_devi)
        # total energy
        with open(p / "log.lammps", "r") as fopen:
            lines = fopen.readlines()
        for idx, line in enumerate(lines):
            if line.startswith('Minimization stats:'):
                stat_idx = idx
                break
        else:
            raise ValueError('error in lammps minimization.')
        en = float(lines[stat_idx+3].split()[-1])
        # natoms
        dump_atoms = read(
            p / "surface.dump", ':', 'lammps-dump-text', 
            specorder=["O","Pt"], units="metal"
        )[-1]
        natoms = len(dump_atoms)
        #print(natoms)
        cur_data = StepInfo(natoms, en, en_devi)
    
        return cur_data
    
    
    data = []
    for p in tqdm(step_dirs_sorted[:10]):
        cur_data = extract_data(p)
        data.append(cur_data)
    
    # write to file
    content = "#{}  {}  {}  {}\n".format("step   ", "natoms  ", "energy  ", "deviation")
    for i, info in enumerate(data):
        content += ("{:8d}  "*2+"{:8.4f}  "*2+"\n").format(
            i, info.natoms, info.energy, info.deviation
        )
    with open("evolution.dat", "w") as fopen:
        fopen.write(content)
    
    exit()
    
    import matplotlib as mpl
    mpl.use('Agg') #silent mode
    from matplotlib import pyplot as plt
    plt.style.use('presentation')
    
    fig, axarr = plt.subplots(
        nrows=1, ncols=1, 
        gridspec_kw={'hspace': 0.3}, figsize=(16,12)
    )
    axarr = [axarr]
    plt.suptitle("GCMC Evolution")
    
    ax = axarr[0]
    # element_numbers = Counter()
    natoms_array = np.array([a.natoms for a in data])
    # oxygen_array = np.array([Counter(a.get_chemical_symbols())["O"] for a in frames])
    steps = range(len(data))
    energies = np.array([a.energy for a in data]) / natoms_array
    en_stdvars = np.array([a.deviation for a in data]) / natoms_array
    
    ax.set_xlabel("MC Step")
    ax.set_ylabel("Energy [eV]")
    
    ax.plot(steps, energies, label="energy per atom")
    apex = 10.
    ax.fill_between(
        steps, energies-apex*en_stdvars, energies+apex*en_stdvars, alpha=0.2,
        label="10 times deviation per atom"
    )
    ax.legend()
    
    plt.savefig("m-stdvar.png")


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np

import jax
import jax.numpy as jnp

from ase.io import read, write

from dscribe.descriptors import ValleOganov

#def compute_pcv(r, r_k, coef):
#    """"""
#    num, dim = r_k.shape
#    #k = np.arange(1, 1+num)
#    k = jnp.arange(0, num)
#
#    #r_norm2 = jnp.linalg.norm(r-r_k, axis=1)**2
#    #r_norm2 = jnp.linalg.norm(r-r_k, axis=1)
#    #r_norm2 = jnp.linalg.norm(r-r_k, axis=1)/dim
#    #r_norm2 = jnp.linalg.norm(r-r_k, axis=1)/jnp.sqrt(dim)
#    r_norm2 = jnp.sum(r*r_k, axis=1)/jnp.linalg.norm(r)/jnp.linalg.norm(r_k, axis=1)
#
#    r_exp = jnp.exp(-coef*r_norm2)
#
#    #print(f"r_norm2: {r_norm2}")
#    #print(f"r_exp: {r_exp}")
#
#    s = (jnp.sum(k*r_exp)) / jnp.sum(r_exp)
#    z = -1./coef*jnp.log(jnp.sum(r_exp))
#
#    ret = jnp.hstack([s, z])
#
#    return ret, ret

def compute_pcv(r, r_k, coef):
    """"""
    num, dim = r_k.shape
    #k = np.arange(1, 1+num)
    k = jnp.arange(0, num)

    #r_norm2 = np.linalg.norm(r-r_k, axis=1)**2
    #r_norm2 = np.linalg.norm(r-r_k, axis=1)
    #r_norm2 = np.linalg.norm(r-r_k, axis=1)/dim
    #r_norm2 = np.linalg.norm(r-r_k, axis=1)/np.sqrt(dim)
    #r_norm2 = np.sum((r-r_k)**2, axis=1)/dim**2
    r_norm2 = jnp.sum(r*r_k, axis=1)/jnp.linalg.norm(r)/jnp.linalg.norm(r_k, axis=1)

    r_exp = jnp.exp(-coef*r_norm2)

    #print(f"r_norm2: {r_norm2}")
    #print(f"r_exp: {r_exp}")

    s = (jnp.sum(k*r_exp)) / jnp.sum(r_exp)
    z = -1./coef*jnp.log(jnp.sum(r_exp))

    ret = jnp.hstack([s, z])

    return ret, ret


class FingerprintColvar():

    def __init__(self, data, *args, **kwargs) -> None:
        """"""

        # - 
        self.params = {}

        # - landmarks
        frames_ = read(data, ":")
        frames = [a[240:] for a in frames_]

        _, self.features = self.calculate_features(frames)

        self.coefficient = 20.

        return

    @property
    def dim(self):
        """"""

        return 2

    def calculate_features(self, frames):
        """"""
        vo2 = ValleOganov(
            species = ["Cu"],
            function = "distance",
            sigma = 10**(-5),
            n = 161,
            #n = 17,
            r_cut = 8.,
        )

        vo3 = ValleOganov(
            species = ["Cu"],
            function = "angle",
            sigma = 10**(-5),
            n = 181,
            r_cut = 4.,
        )

        #vo2_features = vo2.create(frames)
        vo2_gradients, vo2_features = vo2.derivatives(frames)
        #print(vo2_features.shape)
        #print(vo2_gradients.shape)

        return vo2_gradients, vo2_features
    
    def cvfunc(self, atoms, params):
        """"""
        gradients, features = self.calculate_features([atoms[240:]])
        features = features[np.newaxis, :]
        gradients = gradients[np.newaxis, :, :, :]
        #print(f"feature shape: {features.shape}")
        #print(f"gradient shape: {gradients.shape}")
        #print(features)
        #print(self.features)

        curr_features = features
        mark_features = self.features

        #curr_features = jnp.array(features, dtype=jnp.float64)
        #mark_features = jnp.array(self.features, dtype=jnp.float64)

        #cv, _ = compute_pcv(curr_features, mark_features, self.coefficient)
        #print(cv)

        cvg, cv = jax.jacfwd(compute_pcv, argnums=0, has_aux=True)(
            curr_features, mark_features, self.coefficient
        )
        cv = cv[jnp.newaxis, :]
        #print(cv)
        #print(cvg)

        return cv


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

import os
import pathlib
from typing import NoReturn

from . import AbstractPotentialManager


class XtbManager(AbstractPotentialManager):

    name = "xtb"

    implemented_backends = ["xtb"]
    valid_combinations = (
        ("xtb", "ase")
    )

    """See XTB documentation for calculator parameters.

    method: "GFN2-xTB"
    accuracy: 1.0
    electronic_temperature: 300.0
    max_iterations: 250
    solvent: "none"
    cache_api: True

    """

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return
    
    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        try:
            from xtb.ase.calculator import XTB
        except:
            print("Please install xtb python to use this module.")
            exit()

        if self.calc_backend == "xtb":
            calc_cls = XTB
        else:
            raise NotImplementedError(f"Unsupported backend {self.calc_backend}.")
        
        calc = calc_cls(**calc_params)

        self.calc = calc

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

from ase.io import read
from ase.constraints import FixAtoms
from ase.calculators.emt import EMT
from ase.neb import NEB
#from ase.neb import SingleCalculatorNEB
from ase.optimize import BFGS
from ase.io import read, write

from eann.interface.ase.calculator import Eann

#initial = read('initial.traj')
#final = read('final.traj')

#prepared_images = read(
#    "/users/40247882/scratch2/validations/structures/surfaces-2x2/relaxed/hcp2fcc-brg_opt.xyz", ":" # read neb optimised trajectory
#)
#
#for atoms in prepared_images:
#    atoms.wrap()
#write("wrapped_traj.xyz", prepared_images)
#exit()

#prepared_images = read("./start_images.xyz", ":")
prepared_images = read("./hcp2fcc-top_opt.xyz", ":")

initial = prepared_images[0].copy()
final = prepared_images[-1].copy()

constraint = FixAtoms(indices=[1,2,3,4]) # first atom is O
#constraint = FixAtoms(indices=[1,2,3,4,5,6,7,8]) # first atom is O

nimages = 7
images = [initial]
images += [initial.copy() for i in range(nimages-2)]
images.append(final)

# set calculator
cur_model = "/users/40247882/scratch2/oxides/eann-main/reduce-13/ensemble/model-0/eann_best_DOUBLE.pt"
calc = Eann(
    type_map = {'O': 0, 'Pt': 1},
    model = cur_model
)

for atoms in images:
    calc = Eann(
        type_map = {'O': 0, 'Pt': 1},
        #model = "/users/40247882/projects/oxides/eann-main/it-0012/ensemble/model-3/eann_best-3_DOUBLE.pt"
        model = cur_model
    )
    atoms.calc = calc
    atoms.set_constraint(constraint)

print(initial.get_potential_energy())
#print(initial.get_forces())
print(final.get_potential_energy())

neb = NEB(
    images, allow_shared_calculator=False,
    k=0.1
    # dynamic_relaxation = False
)
#neb = SingleCalculatorNEB(images)

neb.interpolate()
#print(neb.images)

qn = BFGS(neb, trajectory="neb.traj")
qn.run(fmax=0.05, steps=50)

opt_images = read("./neb.traj", "-%s:" %nimages)
for a in opt_images:
    calc = Eann(
        type_map = {'O': 0, 'Pt': 1},
        model = cur_model
    )
    a.calc = calc

energies = np.array(
    [a.get_potential_energy() for a in opt_images]
)
energies = energies - energies[0]
print(energies)

if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import dataclasses
import os
import re
import pathlib
import traceback

from typing import Union, List

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.vasp import Vasp

from .. import read_sort, resort_atoms_with_spc, run_ase_calculator
from .string import AbstractStringReactor, StringReactorSetting


#: Ase-vasp sort fname.
ASE_VASP_SORT_FNAME: str = "ase-sort.dat"


def read_vaspout(
    lines: List[str],
) -> int:
    """"""
    pattern = re.compile("[0-9]+ F=")

    steps = [0]
    for line in lines:
        m = pattern.match(line.strip())
        if m:
            step = int(m.group().split()[0])
            steps.append(step)
        else:
            ...
    # print(f"{steps =}")

    return max(steps)


@dataclasses.dataclass
class VaspStringReactorSetting(StringReactorSetting):

    backend: str = "vasp"

    #: Number of tasks/processors/cpus for each image.
    ntasks_per_image: int = 1

    def __post_init__(self):
        """"""
        self._internals.update(
            # ---
            ibrion=3,
            potim=0,
            isif=2,
            # ---
            lclimb=self.climb,
            ichain=0,
            images=self.nimages - 2,
            iopt=1,
            spring=-5,
        )

        return

    def get_run_params(self, *args, **kwargs):
        """"""
        # - convergence criteria
        fmax_ = kwargs.get("fmax", self.fmax)
        steps_ = kwargs.get("steps", self.steps)

        run_params = dict(
            constraint=kwargs.get("constraint", self.constraint),
            ediffg=fmax_ * -1.0,
            nsw=steps_,
        )

        return run_params


class VaspStringReactor(AbstractStringReactor):

    name: str = "vasp"

    traj_name: str = "01/OUTCAR"

    def __init__(
        self,
        calc: Vasp = None,
        params: dict = {},
        ignore_convergence: bool = False,
        directory: Union[str, pathlib.Path] = "./",
        *args,
        **kwargs,
    ) -> None:
        """"""
        self.calc = calc
        if self.calc is not None:
            self.calc.reset()

        self.ignore_convergence = ignore_convergence

        self.directory = directory

        # - parse params
        self.setting = VaspStringReactorSetting(**params)
        self._debug(self.setting)

        return

    def _verify_checkpoint(self):
        """Check if the current directory has any valid outputs or it just created
        the input files.

        """
        verified = super()._verify_checkpoint()
        if verified:
            vasprun = self.directory / "01" / "OUTCAR"
            if vasprun.exists() and vasprun.stat().st_size != 0:
                temp_frames = read(vasprun, ":")
                try:
                    _ = temp_frames[0].get_forces()
                    verified = True
                except:
                    verified = False
            else:
                verified = False
        else:
            verified = False

        return verified

    def _irun(self, structures: List[Atoms], ckpt_wdir=None, *args, **kwargs):
        """"""
        try:
            # --
            run_params = self.setting.get_run_params(**kwargs)
            run_params.update(**self.setting.get_init_params())

            if ckpt_wdir is None:  # start from the scratch
                self._print("interpolate input images...")
                images = self._align_structures(structures, run_params)
            else:
                self._print("update input images...")
                # - update structures
                rep_dirs = sorted(
                    ckpt_wdir.glob(r"[0-9][0-9]"), key=lambda x: int(x.name)
                )

                frames_ = []
                for x in rep_dirs[1:-1]:
                    frames_.append(read(x / "OUTCAR", ":"))
                nframes_per_image = [len(x) for x in frames_]
                nframes = min(nframes_per_image)
                assert nframes > 0, "At least one step finished before resume..."
                intermediates_ = [x[nframes - 1] for x in frames_]

                # -- sort frames from outcar
                if (ckpt_wdir / ASE_VASP_SORT_FNAME).exists():
                    sort, resort = read_sort(ckpt_wdir, ASE_VASP_SORT_FNAME)
                else:
                    natoms = len(structures[0])
                    sort, resort = list(range(natoms)), list(range(natoms))

                intermediates = []
                for a in intermediates_:
                    sorted_atoms = resort_atoms_with_spc(
                        a, resort, "vasp", print_func=self._print, debug_func=self._debug
                    )
                    intermediates.append(sorted_atoms)

                images = [structures[0]] + intermediates + [structures[-1]]

                # the param keys have been proprocessed to vasp ones
                run_params.update(nsw=self.setting.steps + 1 - nframes)

                # constraint info has already been in OUTCAR
                run_params.pop("constraint")

            write(self.directory / "images.xyz", images)

            # - update input
            self.calc.set(**run_params)

            atoms = images[0]
            atoms.calc = self.calc

            # -- write input files
            self.calc.write_input(atoms)
            if (self.directory / "POSCAR").exists():
                os.remove(self.directory / "POSCAR")

            # -- add replica information
            for i, a in enumerate(images):
                rep_dir = self.directory / str(i).zfill(2)
                # It has already been created when images are written.
                # If the previous run has no outputs, we just overwrite everything.
                rep_dir.mkdir(exist_ok=True)
                write(
                    rep_dir / "POSCAR",
                    a[self.calc.sort],
                    symbol_count=self.calc.symbol_count,
                )

            # - run calculation
            run_ase_calculator("vasp", atoms.calc.command, self.directory)

        except Exception as e:
            self._debug(e)
            self._debug(traceback.print_exc())

        return

    def read_convergence(self, *args, **kwargs):
        """Check whether vasp-neb is converged.

        The convergence meets when either the required force is reached or
        the maximum steps exceed.

        """
        converged = super().read_convergence(*args, **kwargs)

        self._print(f"{self.directory =}")
        vaspout_fpath = self.directory / "vasp.out"
        if vaspout_fpath.exists():
            with open(self.directory / "vasp.out", "r") as fopen:
                lines = fopen.readlines()

            steps = read_vaspout(lines)
            if steps >= self.setting.steps:
                converged = True

            for line in lines:
                if "reached required accuracy" in line:
                    converged = True
                    break
        else:
            ...

        return converged

    def _read_a_single_trajectory(self, wdir, *args, **kwargs):
        """

        NOTE: Fixed atoms have zero forces.

        """
        self._debug(f"***** read_trajectory *****")
        self._debug(f"{str(wdir)}")

        images = read(wdir / "images.xyz", ":")
        ini_atoms, fin_atoms = images[0], images[-1]

        # TODO: energy and forces of IS and FS?
        calc = SinglePointCalculator(
            ini_atoms,
            energy=ini_atoms.info["energy"],
            forces=np.zeros((len(ini_atoms), 3)),
        )
        ini_atoms.calc = calc
        calc = SinglePointCalculator(
            fin_atoms,
            energy=fin_atoms.info["energy"],
            forces=np.zeros((len(fin_atoms), 3)),
        )
        fin_atoms.calc = calc

        # - read OUTCARs
        if (self.directory / ASE_VASP_SORT_FNAME).exists():
            sort, resort = read_sort(self.directory, ASE_VASP_SORT_FNAME)
        else:
            natoms = len(ini_atoms)
            sort, resort = list(range(natoms)), list(range(natoms))

        frames_ = []
        for i in range(1, self.setting.nimages - 1):
            curr_frames = read(wdir / f"{str(i).zfill(2)}" / "OUTCAR", ":")
            sorted_frames = []
            for a in curr_frames:
                sorted_atoms = resort_atoms_with_spc(
                    a, resort, "vasp", print_func=self._print, debug_func=self._debug
                )
                sorted_frames.append(sorted_atoms)
            frames_.append(sorted_frames)

        # nframes may not consistent across replicas
        # due to unfinished calculations
        nframes_list = [len(x) for x in frames_]
        nsteps = min(nframes_list)

        frames = []
        for i in range(nsteps):
            curr_frames = (
                [ini_atoms]
                + [frames_[j][i] for j in range(self.setting.nimages - 2)]
                + [fin_atoms]
            )
            frames.append(curr_frames)

        return frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import NoReturn, Union

class Placeholder: # Placeholder

    """Placeholder for input structures that may be from external files.
    """

    #: Working directory for the operation.
    _directory: Union[str,pathlib.Path] = pathlib.Path.cwd()

    #: Working status that should be always finished.
    status = "finished"

    def __init__(self):
        """"""
        self.consumers = []

        return

    @property
    def directory(self):
        """"""

        return self._directory
    
    @directory.setter
    def directory(self, directory_) -> NoReturn:
        """"""
        self._directory = pathlib.Path(directory_)

        return

    def reset(self):
        """Reset node's output and status."""
        ...

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings
import pathlib
from pathlib import Path
from typing import NoReturn, List, Union

from ..core.register import registers

from .dimer import DimerValidator
registers.validator.register("dimer")(DimerValidator)

from .mdf import MassDistributionValidator
registers.validator.register("mass_distribution")(MassDistributionValidator)

from .rdf import RdfValidator
registers.validator.register("radial_distribution")(RdfValidator)

from .eos import EquationOfStateValidator
registers.validator.register("equation_of_state")(EquationOfStateValidator)

from .melting_point import MeltingPointValidator
registers.validator.register("melting_point")(MeltingPointValidator)

from .minima import MinimaValidator
registers.validator.register("minima")(MinimaValidator)

from .spc import SinglepointValidator
registers.validator.register("spc")(SinglepointValidator)

from .surface_energy import SurfaceEnergyValidator
registers.validator.register("surface_energy")(SurfaceEnergyValidator)

from .diffusion_coefficient import DiffusionCoefficientValidator
registers.validator.register("diffusion_coefficient")(DiffusionCoefficientValidator)

from .rank import RankValidator
registers.validator.register("rank")(RankValidator)

from .rxn import PathwayValidator
registers.validator.register("mep")(PathwayValidator)


"""
Various properties to be validated

Atomic Energy and Crystal Lattice constant

Elastic Constants

Phonon Calculations

Point Defects (vacancies, self interstitials, ...)

Surface energies

Diffusion Coefficient

Adsorption, Reaction, ...
"""


def run_validation(params: dict, directory: Union[str, pathlib.Path], potter):
    """ This is a factory to deal with various validations...
    """
    # run over validations
    directory = pathlib.Path(directory)

    raise NotImplementedError("Command Line Validation is NOT Suppoted.")


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
import os
import pathlib

import numpy as np

from reportlab.platypus import SimpleDocTemplate, Image, Table, Paragraph, PageBreak

from ase.io import read, write

from . import registers
from .selector import AbstractSelector


class CompareSelector(AbstractSelector):

    name: str = "compare"

    default_parameters: dict = dict(comparator_name=None, comparator_params={})

    def __init__(self, directory="./", axis=None, *args, **kwargs) -> None:
        """"""
        super().__init__(directory, axis, *args, **kwargs)

        self.comparator = registers.create(
            "comparator",
            self.comparator_name,
            convert_name=True,
            **self.comparator_params,
        )

        return

    def _mark_structures(self, data, *args, **kwargs) -> None:
        """"""
        super()._mark_structures(data, *args, **kwargs)

        # -
        structures = data.get_marked_structures()
        nstructures = len(structures)

        # - start from the first structure and compare structures by a given comparator
        if not hasattr(self.comparator, "prepare_data"):
            selected_indices, scores = [0], []
            for i, a1 in enumerate(structures[1:]):
                # NOTE: assume structures are sorted by energy
                #       close structures may have a high possibility to be similar
                #       so we compare reversely
                for j in selected_indices[::-1]:
                    self._print(f"compare: {i+1} and {j}")
                    a2 = structures[j]
                    if self.comparator(a1, a2):
                        break
                else:
                    selected_indices.append(i + 1)
                    self._print(f"--->>> current indices: {selected_indices}")
        else:
            fingerprints = self.comparator.prepare_data(structures)

            # - compare structural fingerprint
            selected_indices, unique_groups, scores = [], {}, []
            for i, fp in enumerate(fingerprints):
                for j in selected_indices[::-1]:
                    # TODO: The ditribution maybe too wide?
                    fp_avg = np.average(
                        [fingerprints[x] for x in unique_groups[j]], axis=0
                    )
                    if self.comparator(fp, fp_avg):
                        unique_groups[j].append(i)
                        break
                else:
                    selected_indices.append(i)
                    unique_groups[i] = [i]

            # - merge groups with similar average fingerprint
            unique_fingerprints = {}
            for k, v in unique_groups.items():
                unique_fingerprints[k] = np.average(
                    [fingerprints[x] for x in v], axis=0
                )

            selected_group_indices, merged_groups = [], {}
            for i, fp in unique_fingerprints.items():
                for j in selected_group_indices:
                    # TODO: The ditribution maybe too wide?
                    fp_avg = np.average(
                        [unique_fingerprints[x] for x in merged_groups[j]], axis=0
                    )
                    if self.comparator(fp, fp_avg):
                        merged_groups[j].append(i)
                        break
                else:
                    selected_group_indices.append(i)
                    merged_groups[i] = [i]

            new_selected_indices, new_unique_groups = [], {}
            for k, v in merged_groups.items():
                curr_indices = sorted(
                    list(itertools.chain(*[unique_groups[x] for x in v]))
                )
                # print(curr_indices)
                # print(unique_fingerprints[k])
                new_unique_groups[curr_indices[0]] = curr_indices
                new_selected_indices.append(curr_indices[0])
            selected_indices = new_selected_indices
            unique_groups = new_unique_groups

            # - report data
            unique_wdir = self.directory / "unique"
            unique_wdir.mkdir(parents=True, exist_ok=True)

            # -- save structures per group?
            with open(unique_wdir / "unique.dat", "w") as fopen:
                for k, v in unique_groups.items():
                    fopen.write(f"{k}_{len(v)}: {v}\n")
            #        write(unique_wdir/f"g_{k}.xyz", [structures[i] for i in v])

            # -- write a report?
            story = []
            for i, (k, v) in enumerate(unique_groups.items()):
                curr_frames = [structures[x] for x in v]
                curr_nframes = len(curr_frames)
                if curr_nframes > 1:
                    story.append(Paragraph(f"Group {k}"))
                    story.append(Paragraph(" ".join([str(x) for x in v])))
                    images = []
                    for ia, a in enumerate(curr_frames[:]):
                        write(unique_wdir / f"xxx-{i}_{ia}.png", a)
                        image = Image(unique_wdir / f"xxx-{i}_{ia}.png")
                        image.drawWidth = 108
                        image.drawHeight = 100
                        images.append(image)
                    image_table = []
                    nrows = int(curr_nframes / 4.0)
                    for x in range(nrows):
                        image_table.append(
                            [images[im] for im in range(x * 4, (x + 1) * 4)]
                        )
                    if (curr_nframes - nrows * 4) > 0:
                        image_table.append(
                            [images[im] for im in range(nrows * 4, curr_nframes)]
                        )

                    story.append(Table(image_table))
                    story.append(PageBreak())

            doc = SimpleDocTemplate(str(self.directory / "report.pdf"))
            doc.build(story)

            for x in unique_wdir.glob("*.png"):
                os.remove(x)

        curr_markers = data.markers
        # NOTE: convert to np.array as there may have 2D markers
        selected_markers = np.array([curr_markers[i] for i in selected_indices])
        data.markers = selected_markers

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy 
from typing import List

from .monte_carlo import MonteCarlo


class BasinHopping(MonteCarlo):

    def __init__(
        self, builder: dict, operators: List[dict], 
        convergence: dict, random_seed=None, dump_period: int=1, ckpt_period: int=100,
        restart: bool = False, directory="./", *args, **kwargs
    ) -> None:
        """"""
        super().__init__(
            builder, operators, convergence, random_seed, dump_period, ckpt_period,
            restart, directory, *args, **kwargs
        )

        return
    
    def as_dict(self) -> dict:
        """"""
        engine_params = super().as_dict()
        engine_params["method"] = "basin_hopping"

        return engine_params


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import pathlib
import time

from typing import NoReturn, Union, Tuple, List, Callable


from ..placeholder import Placeholder
from ..variable import Variable
from .session import AbstractSession
from .utils import traverse_postorder


class ActiveSession(AbstractSession):

    def __init__(
        self,
        steps: int = 2,
        reset_random_state: bool = False,
        reset_random_config: Tuple[str, int] = ("init", 0),
        directory="./",
    ) -> None:
        """Initialise an ActiveSession.

        Args:
            steps: Number of active learning steps.
            reset_random_seed: A tuple of a str and a int.

        """
        self.steps = steps

        # - random-related ...
        self.reset_random_state = reset_random_state
        assert reset_random_config[0] in [
            "init",
            "zero",
        ], "Reset random seed mode must either be init or zero."
        self.reset_random_seed_mode = reset_random_config[0]
        self.reset_random_seed_step = reset_random_config[1]

        self.directory = pathlib.Path(directory)

        return

    def run(self, operation, feed_dict: dict = {}, *args, **kwargs) -> None:
        """"""
        self.state = "StepToStart"
        # - update nodes' attrs based on the previous iteration
        # nodes_postorder = traverse_postorder(operation)
        # for node in nodes_postorder:
        #    if hasattr(node, "enable_active"):
        #        node.enable_active()
        if self.reset_random_state:
            self._print(
                f"RESET RANDOM SEED - MODE: {self.reset_random_seed_mode} STEP: {self.reset_random_seed_step}"
            )

        # -
        for curr_step in range(self.steps):
            curr_wdir = self.directory / f"iter.{str(curr_step).zfill(4)}"
            # -- run operation
            nodes_postorder = traverse_postorder(operation)
            # NOTE:
            if self.reset_random_state and curr_step >= self.reset_random_seed_step:
                for node in nodes_postorder:
                    if hasattr(node, "reset_random_seed"):
                        self._print(f"reset {node.directory.name}'s random seeds.")
                        node.reset_random_seed(mode=self.reset_random_seed_mode)
            # -- run operations
            self._irun(
                wdir=curr_wdir,
                nodes_postorder=nodes_postorder,
                feed_dict=feed_dict,
            )
            if not (self.state == "StepFinished"):
                self._print("wait current iteration to finish...")
            else:
                # NOTE: If previous step finished, the nodes may not have outputs
                #       as we skip them...
                if not (curr_wdir / "FINISHED").exists():
                    # --- report
                    self._print("[{:^24s}]".format("CONVERGENCE"))
                    converged_list = []
                    for node in nodes_postorder:
                        if hasattr(node, "report_convergence"):
                            converged = node.report_convergence()
                            converged_list.append(converged)
                    if converged_list and all(converged_list):
                        self._print(f"Active Session converged at step {curr_step}.")
                        self.state = "LoopConverged"
                    else:
                        self._print(f"Active Session UNconverged at step {curr_step}.")
                        if curr_step + 1 == self.steps:
                            self.state = "LoopUnConverged"
                        else:
                            ...  # Just StepFinished
                    # --- save state to FILE
                    with open(curr_wdir / "FINISHED", "w") as fopen:
                        fopen.write(
                            f"STATE {self.state} FINISHED AT {time.asctime( time.localtime(time.time()) )}."
                        )
                else:
                    self._print(
                        "[{:^24s}] FINISHED".format(f"STEP.{str(curr_step).zfill(4)}")
                    )
            # -- add an atrribute that indicates all steps are finished
            if self.state != "StepFinished":
                break
        else:
            self.state = "LoopFinished"

        return

    def _irun(self, wdir, nodes_postorder, feed_dict: dict = {}) -> None:
        """"""
        if (wdir / "FINISHED").exists():
            self.state = "StepFinished"
            return

        # - clear previous nodes' outputs
        #   somtimes two steps run consecutively and some nodes in the second step
        #   breaks and make its following nodes use outputs from the last step,
        #   which is a hidden error
        for node in nodes_postorder:
            node.reset()

        # - find forward order
        self._print(
            "[{:^24s}] NUM_NODES: {} AT MAIN: {}".format(
                "START", len(nodes_postorder), str(wdir)
            )
        )

        # - run nodes
        self.state = "StepFinished"
        for i, node in enumerate(nodes_postorder):
            # -- change version ...
            if hasattr(node, "version"):
                node.version = wdir.name

            # NOTE: reset directory since it maybe changed
            prev_name = node.directory.name.split(".")[-1]  # remove previous orders
            if not prev_name:
                prev_name = node.__class__.__name__
            # prev_name = node.__class__.__name__
            node.directory = wdir / f"{str(i).zfill(4)}.{prev_name}"
            if node.__class__.__name__.endswith("Variable"):
                node_type = "VX"
            else:
                node_type = "OP"
            self._print(
                "[{:^24s}] NAME: {} AT {}".format(
                    node_type, node.__class__.__name__.upper(), node.directory.name
                )
            )

            # -- forward
            if isinstance(node, Placeholder):
                node.output = feed_dict[node]
            elif isinstance(node, Variable):
                node.output = node.value
            else:  # Operation
                # FIXME: If the session has many branches,
                #        how do we define the state?
                self._debug(f"node: {node}")
                self._process_operation(node)

        return


class OTFSession(AbstractSession):

    def __init__(self, directory="./") -> None:
        """"""
        self.directory = pathlib.Path(directory)

        return

    def run(self, operation, feed_dict: dict = {}) -> None:
        """"""
        for curr_step in range(3):
            # -- update operation
            if curr_step > 0:
                curr_step = curr_step
                self._update_potter(operation)
            # -- run operation
            finished = self._irun(
                operation=operation,
                feed_dict=feed_dict,
                wdir=self.directory / f"iter.{str(curr_step).zfill(4)}",
            )
            if not finished:
                self._print("wait current iteration to finish...")
                break
        else:
            ...  # ALL iterations finished...

        return

    def _update_potter(self, operation):
        """"""

        return

    def _irun(self, operation, feed_dict: dict = {}, wdir=None):
        """"""
        # - find forward order
        nodes_postorder = traverse_postorder(operation)
        self._print(
            "[{:^24s}] NUM_NODES: {} AT MAIN: {}".format(
                "START", len(nodes_postorder), str(wdir)
            )
        )

        # - run nodes
        finished = True
        for i, node in enumerate(nodes_postorder):
            # NOTE: reset directory since it maybe changed
            # prev_name = node.directory.name
            # if not prev_name:
            #    prev_name = node.__class__.__name__
            prev_name = node.__class__.__name__
            node.directory = wdir / f"{str(i).zfill(4)}.{prev_name}"
            if node.__class__.__name__.endswith("Variable"):
                node_type = "VX"
            else:
                node_type = "OP"
            self._print(
                "[{:^24s}] NAME: {} AT {}".format(
                    node_type, node.__class__.__name__.upper(), node.directory.name
                )
            )

            if isinstance(node, Placeholder):
                node.output = feed_dict[node]
            elif isinstance(node, Variable):
                node.output = node.value
            else:  # Operation
                self._debug(f"node: {node}")
                if node.is_ready_to_forward():
                    node.inputs = [input_node.output for input_node in node.input_nodes]
                    node.output = node.forward(*node.inputs)
                else:
                    finished = False
                    self._print("wait previous nodes to finish...")
                    continue

        return finished


class CyclicSession(AbstractSession):
    """Create a cyclic session.

    This supports a session that contains preprocess, iteration, and postprocess.

    """

    def __init__(self, directory="./") -> None:
        """"""
        self.directory = pathlib.Path(directory)

        return

    def iteration(self, operations):
        """Iterative part that converges at certain criteria.

        This noramlly includes steps: sample, select, label, and train. Some inputs of
        operations should be update during the iterations, for instance, the models
        in the potential.

        """

        return

    def run(
        self, init_node, iter_node=None, post_node=None, repeat=1, *args, **kwargs
    ) -> None:
        """"""
        # - init
        self._print(("=" * 28 + "{:^24s}" + "=" * 28 + "\n").format("INIT"))
        session = Session(self.directory / "init")
        _ = session.run(init_node, feed_dict={})
        self._print("status: ", init_node.status)

        init_state = init_node.status
        if init_state != "finished":
            self._print("wait init session to finish...")
            return

        # - iter
        curr_potter_node = init_node  # a node that forwards a potter manager
        for i in range(repeat):
            self._print(
                ("=" * 28 + "{:^24s}" + "=" * 28 + "\n").format(
                    f"ITER.{str(i).zfill(4)}"
                )
            )
            session = Session(self.directory / f"iter.{str(i).zfill(4)}")
            # -- update some parameters
            curr_node = copy.deepcopy(iter_node)
            nodes_postorder = traverse_postorder(curr_node)
            # --- trainer
            # --- potter
            for x in nodes_postorder:
                if x.__class__.__name__ == "drive":
                    x.input_nodes[1]._update_workers(curr_potter_node)
            # -- run
            _ = session.run(curr_node, feed_dict={})
            curr_state = curr_node.status
            if curr_state != "finished":
                self._print(f"wait iter.{str(i).zfill(4)} session to finish...")
                break
            else:
                curr_potter_node = curr_node
        else:
            self._print(f"iter finished...")

        # - post

        return

    def _irun(self, operation, feed_dict: dict = {}) -> NoReturn:
        """"""
        # - find forward order
        nodes_postorder = traverse_postorder(operation)
        self._print(
            "[{:^24s}] NUM_NODES: {} AT MAIN: {}".format(
                "START", len(nodes_postorder), str(self.directory)
            )
        )

        # - run nodes
        for i, node in enumerate(nodes_postorder):
            # NOTE: reset directory since it maybe changed
            prev_name = node.directory.name
            if not prev_name:
                prev_name = node.__class__.__name__
            node.directory = self.directory / f"{str(i).zfill(4)}.{prev_name}"
            if node.__class__.__name__.endswith("Variable"):
                node_type = "VX"
            else:
                node_type = "OP"
            self._print(
                "[{:^24s}] {:<36s} AT {}".format(
                    node_type, node.__class__.__name__.upper(), node.directory.name
                )
            )

            if isinstance(node, Placeholder):
                node.output = feed_dict[node]
            elif isinstance(node, Variable):
                node.output = node.value
            else:  # Operation
                if node.is_ready_to_forward():
                    node.inputs = [input_node.output for input_node in node.input_nodes]
                    node.output = node.forward(*node.inputs)
                else:
                    print("wait previous nodes to finish...")
                    continue

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import subprocess
from pathlib import Path
from itertools import groupby
from operator import itemgetter

from typing import Any, Union, List

import json
import yaml


def dict2str(d: dict, indent: int = 2):
    """Convert a nested dict to str."""

    def _dict2str(d_: dict, indent_: int):
        """Recursive function."""
        content = ""
        for k, v in d_.items():
            if isinstance(v, dict):
                content += f"{k}:\n" + _dict2str(v, indent_ + indent)
            else:
                content += " " * indent_ + f"{k}: {v}\n"

        return content

    content = _dict2str(d, 0)

    return content


class CustomTimer:

    def __init__(self, name="code", func=print):
        """"""
        self.name = name
        self._print = func

        return

    def __call__(self, func) -> Any:
        """"""

        def func_timer(*args, **kwargs):
            st = time.time()
            ret = func(*args, **kwargs)
            et = time.time()
            content = (
                "*** " + self.name + " time: " + "{:>8.4f}".format(et - st) + " ***"
            )
            self._print(content)

            return ret

        return func_timer

    def __enter__(self):
        """"""
        self.st = time.time()  # start time

        return self

    def __exit__(self, *args):
        """"""
        self.et = time.time()  # end time

        content = (
            "*** "
            + self.name
            + " time: "
            + "{:>8.4f}".format(self.et - self.st)
            + " ***"
        )
        self._print(content)

        return


def check_path(target_dir: Union[str, Path]) -> bool:
    """check path existence, if so skip the following
    TODO: add output option
    make this into a context?
    """
    target_dir = Path(target_dir)
    if not target_dir.exists():
        target_dir.mkdir(parents=True)
    else:
        print(f"  {target_dir.name} exists, so next...")

    return


def find_backups(dpath, fname, prefix="bak"):
    """find a series of files in a dir
    such as fname, bak.0.fname, bak.1.fname
    """
    dpath = Path(dpath)
    fpath = dpath / fname
    if not fpath.exists():
        raise FileNotFoundError(f"fpath does not exist.")

    backups = list(dpath.glob(prefix + ".[0-9]*." + fname))
    backups = sorted(backups, key=lambda x: int(x.name.split(".")[1]))
    backups.append(fpath)

    return backups


def run_command(directory, command, comment="", timeout=None):
    proc = subprocess.Popen(
        command,
        shell=True,
        cwd=directory,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    if timeout is None:
        errorcode = proc.wait()
    else:
        errorcode = proc.wait(timeout=timeout)

    msg = "Message: " + "".join(proc.stdout.readlines())
    print(msg)
    if errorcode:
        raise RuntimeError("Error in %s at %s." % (comment, directory))

    return msg


def convert_indices(indices: Union[str, List[int]], index_convention="lmp"):
    """parse indices for reading xyz by ase, get start for counting
    constrained indices followed by lammps convention
    "2:4 3:8"
    convert [1,2,3,6,7,8] to "1:3 6:8"
    lammps convention starts from 1 and includes end
    ---
    input can be either py or lmp
    output for indices is in py since it can be used to access atoms
    output for text is in lmp since it can be used in lammps or sth
    """
    ret = []
    if isinstance(indices, str):
        # string to List[int]
        for x in indices.strip().split():
            cur_range = list(map(int, x.split(":")))
            if len(cur_range) == 1:
                start, end = cur_range[0], cur_range[0]
            else:
                start, end = cur_range
            if index_convention == "lmp":
                ret.extend([i - 1 for i in list(range(start, end + 1))])
            elif index_convention == "py":
                ret.extend(list(range(start, end)))
            else:
                pass
    elif isinstance(indices, list):
        # List[int] to string
        indices = sorted(indices)
        if index_convention == "lmp":
            pass
        elif index_convention == "py":
            indices = [i + 1 for i in indices]
        ret = []
        # ranges = []
        for k, g in groupby(enumerate(indices), lambda x: x[0] - x[1]):
            group = map(itemgetter(1), g)
            group = list(map(int, group))
            # ranges.append((group[0],group[-1]))
            if group[0] == group[-1]:
                ret.append(str(group[0]))
            else:
                ret.append("{}:{}".format(group[0], group[-1]))
        ret = " ".join(ret)
    else:
        pass

    return ret


def parse_input_file(
    input_fpath: Union[str, Path],
    write_json: bool = False,  # write readin dict to check if alright
):
    """"""
    input_dict = None

    # - parse input type
    if isinstance(input_fpath, dict):
        input_dict = input_fpath
        json_path = Path.cwd()
    else:
        if isinstance(input_fpath, str):
            input_file = Path(input_fpath)
            json_path = input_file.parent
        elif isinstance(input_fpath, Path):
            input_file = input_fpath
            json_path = input_file.parent
        else:
            return None

        # --- read dict from files
        try:
            if input_file.suffix == ".json":
                with open(input_file, "r") as fopen:
                    input_dict = json.load(fopen)
            elif input_file.suffix == ".yaml":
                with open(input_file, "r") as fopen:
                    input_dict = yaml.safe_load(fopen)
            else:
                ...
        except FileNotFoundError as e:
            # NOTE: There is json or yaml in the string but it is not a file though.
            input_dict = None

    # NOTE: recursive read internal json or yaml files
    if input_dict is not None:
        for key, value in input_dict.items():
            key_dict = parse_input_file(value, write_json=False)
            if key_dict is not None:
                input_dict[key] = key_dict

    if input_dict and write_json:
        with open(json_path / "params.json", "w") as fopen:
            json.dump(input_dict, fopen, indent=4)
        print("See params.json for values of all parameters...")

    return input_dict


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List

import numpy as np

from ase import Atoms
from ase.data import covalent_radii
from ase.calculators.calculator import Calculator
from ase.neighborlist import NeighborList, natural_cutoffs

from ..utils import compute_distance_and_shift


""""""

#: Argon cluster LJ epsilon.
EPS_AFIR: float = 1.0061 / 96.485

#: Argon cluster LJ radius.
R0_AFIR: float = 3.8164  # Ang

#: Argon cluster LJ power.
POW_AFIR: int = 6


def compute_afir_energy_and_forces(
    positions,
    distances,
    covalent_radii,
    pair_indices,
    pair_shifts,
    *,
    gamma: float = 2.5,
    power: int = POW_AFIR,
    r0: float = R0_AFIR,
    epsilon: float = EPS_AFIR,
):
    """AFIR function in JCTC2011.

    E_afir = \alpha*\frac{\sum_ij ((c_i+c_j)/r_ij)^p*r_ij}{\sum_ij ((c_i+c_j)/r_ij)^p}

    Args:
        pair_indices: shape (num_pairs, 2)
        pair_shifts:  in cartesian coordinates.

    """
    # - compute the prefactor alpha
    alpha = gamma / (
        (2 ** (-1 / 6) - (1 + (1 + gamma / epsilon) ** 0.5) ** (-1 / 6)) * r0
    )

    # - compute pair distances
    # pair_positions = np.take(positions, pair_indices.T, axis=0)
    # dvecs = pair_positions[0] - pair_positions[1] + pair_shifts
    # distances = np.linalg.norm(dvecs, axis=1)

    # - compute inverse distances
    pair_radii = np.take(covalent_radii, pair_indices.T, axis=0)
    weights = ((pair_radii[0] + pair_radii[1]) / distances) ** power

    # - compute energy
    w_s = np.sum(weights)
    dw_s = np.sum(weights * distances)
    energy = alpha * dw_s / w_s

    # - compute forces
    #   w = ((c_i+c_j)/r_ij)^p
    #   dE/dr = ((w+dw*r)*sum_w-sum_(w*r)*dw)/sum_w^2
    #   dE/dr *= alpha
    forces = np.zeros(positions.shape)
    for k, (i, j) in enumerate(pair_indices):
        pos_i, pos_j = positions[i], positions[j] + pair_shifts[k]
        r = distances[k]
        w = weights[k]
        d_w = -power / r * w
        dEdr = alpha * ((w + r * d_w) * w_s - dw_s * d_w) / w_s**2
        frc_ij = -dEdr * (pos_i - pos_j) / r
        forces[i] += frc_ij
        forces[j] -= frc_ij

    return energy, forces


class AFIRCalculator(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    default_parameters = dict(gamma=2.5, power=6, cutoff=6.0)

    def __init__(self, groups, restart=None, label=None, atoms=None, directory=".", **kwargs):
        """"""
        super().__init__(
            restart=restart, label=label, atoms=atoms, directory=directory, **kwargs
        )

        self.gamma = self.parameters["gamma"]  # eV
        assert self.gamma > 0.

        self.radius = self.parameters.get("radius", R0_AFIR)
        self.epsilon = self.parameters.get("epsilon", EPS_AFIR)
        self.power = int(self.parameters.get("power", POW_AFIR))

        self.groups = groups
        self.cutoff = self.parameters["cutoff"]

        assert len(self.groups) == 2

        # ---
        self.neighlist = None

        return

    def calculate(
        self,
        atoms=None,
        properties=["energy"],
        system_changes=["positions", "numbers", "cell"],
    ):
        """"""
        super().calculate(atoms, properties, system_changes)

        # NOTE: check cutoff before assigning groups
        if self.neighlist is None:
            self.neighlist = NeighborList(
                [self.cutoff/2.]*len(atoms),
                skin=0.0, self_interaction=False, bothways=True
            )
        self.neighlist.update(atoms)

        # - get constants
        atomic_numbers = atoms.get_atomic_numbers()
        atomic_radii = np.array([covalent_radii[i] for i in atomic_numbers])

        # - find bond pairs
        bond_pairs, bond_distances, bond_shifts = compute_distance_and_shift(
            atoms, self.neighlist, self.groups[0], self.groups[1]
        )

        # print(f"{bond_pairs =}")
        # print(f"{bond_distances =}")
        # print(f"{bond_shifts =}")

        # - compute properties
        energy, forces = compute_afir_energy_and_forces(
            atoms.positions,
            bond_distances,
            atomic_radii,
            bond_pairs,
            bond_shifts,
            gamma=self.gamma,
            power=self.power,
            r0=self.radius,
            epsilon=self.epsilon,
        )

        # print(f"{energy =}")
        # print(f"{forces =}")

        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


from . import AbstractPotentialManager, DummyCalculator


"""Check https://dftd3.readthedocs.io/en/latest/api/ase.html

To install, use conda install dftd3-python -c conda-forge.

Calculator parameters should have `method` (xc e.g. PBE) and `damping` (e.g. d3bj).

"""


class Dftd3Manager(AbstractPotentialManager):

    name = "dftd3"

    implemented_backends = ["ase"]
    valid_combinations = (
        ("ase", "ase")
    )

    """See ASE documentation for calculator parameters.
    """

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return
    
    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        calc = DummyCalculator()
        if self.calc_backend == "ase":
            from dftd3.ase import DFTD3 as calc_cls
        else:
            raise NotImplementedError(f"Unsupported backend {self.calc_backend}.")
        
        calc = calc_cls(**calc_params)

        self.calc = calc

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List

import numpy as np
from scipy.interpolate import make_interp_spline, BSpline

from ase import Atoms
from ase.geometry import find_mic

def get_properties(frames: List[Atoms], other_props = []):
    """Get properties of frames for comparison.

    Currently, only total energy and forces are considered.

    Returns:
        tot_symbols: shape (nframes,)
        tot_energies: shape (nframes,)
        tot_forces: shape (nframes,3)

    """
    tot_symbols, tot_energies, tot_forces = [], [], []

    for atoms in frames: # free energy per atom
        # -- basic info
        symbols = atoms.get_chemical_symbols()
        tot_symbols.extend(symbols)

        # -- energy
        energy = atoms.get_potential_energy() 
        tot_energies.append(energy)

        # -- force
        forces = atoms.get_forces(apply_constraint=False)
        tot_forces.extend(forces.tolist())

    return tot_symbols, tot_energies, tot_forces

def wrap_traj(frames):
    """Align positions according to the first frame.

    This is necessary for computing physical quantities base on atomic positions 
    with periodic boundary conditions.

    NOTE:
        This only works for fixed cell systems.
    
    TODO:
        Variable cell systems?

    """
    cell = frames[0].get_cell(complete=True)
    nframes = len(frames)
    for i in range(1,nframes):
        prev_positions = frames[i-1].get_positions()
        curr_positions = frames[i].get_positions()
        shift = curr_positions - prev_positions
        curr_vectors, curr_distances = find_mic(shift, cell, pbc=True)
        frames[i].positions = prev_positions + curr_vectors

    return frames

def smooth_curve(bins, points):
    """"""
    spl = make_interp_spline(bins, points, k=3)
    bins = np.linspace(bins.min(), bins.max(), 300)
    points= spl(bins)

    for i, d in enumerate(points):
        if d < 1e-6:
            points[i] = 0.0

    return bins, points


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .insert import GraphInsertModifier
from .remove import GraphRemoveModifier
from .exchange import GraphExchangeModifier


__all__ = [GraphInsertModifier, GraphRemoveModifier, GraphExchangeModifier]


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import collections
import copy
import pathlib

from typing import Union, List, Callable

import numpy as np

from ase import Atoms
from ase.io import read, write

try:
    import dpdata
except:
    ...


def get_formula_from_atoms(atoms: Atoms) -> str:
    """Get atoms' chemical composition sorted by the alphabetic order."""

    chemical_symbols = atoms.get_chemical_symbols()
    composition = collections.Counter(chemical_symbols)
    sorted_composition = sorted(composition.items(), key=lambda x: x[0])

    return "".join([str(k) + str(v) for k, v in sorted_composition])


def convert_groups(
    names: List[str],
    groups: List[List[Atoms]],
    batchsizes: Union[List[int], int],
    type_map: List[str],
    suffix: str,
    dest_dir: Union[str, pathlib.Path] = "./",
    pfunc: Callable = print,
) -> None:
    """Dump structures to dp trainning format."""

    nsystems = len(groups)
    if isinstance(batchsizes, int):
        batchsizes = [batchsizes] * nsystems

    # --- dpdata conversion
    dest_dir = pathlib.Path(dest_dir)
    train_set_dir = dest_dir / f"{suffix}"
    if not train_set_dir.exists():
        train_set_dir.mkdir()

    sys_dirs = []

    cum_batchsizes = 0  # number of batchsizes for training
    for name, frames, batchsize in zip(names, groups, batchsizes):
        nframes = len(frames)
        nbatch = int(np.ceil(nframes / batchsize))
        pfunc(
            f"{suffix} system {name} nframes {nframes} nbatch {nbatch} batchsize {batchsize}"
        )
        # --- check composition consistent
        compositions = [get_formula_from_atoms(a) for a in frames]
        assert (
            len(set(compositions)) == 1
        ), f"Inconsistent composition {len(set(compositions))} =? 1..."
        curr_composition = compositions[0]

        cum_batchsizes += nbatch
        # --- NOTE: need convert forces to force
        frames_ = copy.deepcopy(frames)
        # check pbc
        pbc = np.all([np.all(a.get_pbc()) for a in frames_])
        for atoms in frames_:
            try:
                # NOTE: We need update info and arrays as well
                #       as some dpdata uses data from them instead of calculator
                results = atoms.calc.results
                for k, v in results.items():
                    if k in atoms.info:
                        atoms.info[k] = v
                    if k in atoms.arrays:
                        atoms.arrays[k] = v
                # - change force data
                forces = results["forces"].copy()
                del atoms.arrays["forces"]
                atoms.arrays["force"] = forces
                # - remove some keys as dpdata cannot recognise them
                #   e.g. tags, momenta, initial_charges
                keys = copy.deepcopy(list(atoms.arrays.keys()))
                for k in keys:
                    if k not in ["numbers", "positions", "force"]:
                        del atoms.arrays[k]
            except:
                ...
            finally:
                atoms.calc = None

        # --- convert data
        write(train_set_dir / f"{name}-{suffix}.xyz", frames_)
        dsys = dpdata.MultiSystems.from_file(
            train_set_dir / f"{name}-{suffix}.xyz",
            fmt="quip/gap/xyz",
            type_map=type_map,
        )
        # NOTE: this function create dir with composition and overwrite files
        #       so we need separate dirs...
        sys_dir = train_set_dir / name
        if sys_dir.exists():
            raise FileExistsError(f"{sys_dir} exists. Please check the dataset.")
        else:
            dsys.to_deepmd_npy(train_set_dir / "_temp")  # prec, set_size
            (train_set_dir / "_temp" / curr_composition).rename(sys_dir)
            if not pbc:
                with open(sys_dir / "nopbc", "w") as fopen:
                    fopen.write("nopbc\n")
        sys_dirs.append(sys_dir)
    (train_set_dir / "_temp").rmdir()

    return cum_batchsizes, sys_dirs


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from pathlib import Path

from ase import Atoms
from ase.io import read, write

import ase
from ase.ga.utilities import (closest_distances_generator, atoms_too_close,
                              atoms_too_close_two_sets)

from ase.constraints import FixAtoms

# molecular dynamics
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.langevin import Langevin

from gdpx.computation.aseplumed import AsePlumed


class GridReactionSampling():

    """ steps
        1. slice surface grid
        2. random placement
        3. constrained MD (harmonic potential with large spring constant)
        4. free MD towards IS and FS
        5. collect data and train
    """

    grid_length = 5.0

    constrained_distance = 2.2

    # MD parameters
    temperature = 600 # Kelvin
    timestep = 2.0 # fs

    def __init__(self, reactant, surface):
        """"""
        self.reactant = reactant
        self.surface = surface

        # generate blmin
        # tag every molecules and state atomice types
        #unique_atom_types = []
        #for i, atoms in enumerate(reactants):
        #    atoms.set_tags(i)
        #    unique_atom_types.extend(atoms.get_atomic_numbers())
        #unique_atom_types = list(set(unique_atom_types))
        unique_atom_types = ["C", "O", "Cr", "Zn"]
        unique_atom_types = [6, 8, 24, 30]

        blmin = closest_distances_generator(
            atom_numbers=unique_atom_types,
            ratio_of_covalent_radii=0.8 # be careful with test too far
        )
        self.blmin = blmin

        return
    
    def __partition_surface_grid(self):
        # surface
        surface = self.surface
        positions = surface.positions

        # generate grid
        cell = self.surface.cell.complete()
        print("cell: ", cell)

        a, b, c, alpha, beta, gamma = self.surface.get_cell_lengths_and_angles()

        grid_length = self.grid_length
        na, nb = int(np.round(a/grid_length)), int(np.round(b/grid_length))
        la, lb = a/na, b/nb
        print("number of grids: ", na, nb)
        print("grid size", la, lb)

        # run over regions and place reactant
        candidates = []
        for i in range(na):
            for j in range(nb):
                print("region: ", i, j)
                x1, x2 = i*la, (i+1)*la
                y1, y2 = j*lb, (j+1)*lb
                # find zmax in the region
                indices = []
                for idx, pos in enumerate(positions):
                    if (x1 <= pos[0] < x2) and (y1 <= pos[1] < y2):
                        indices.append(idx)
                print("number in this region: ", len(indices))
                zmax = np.max(positions[indices][:, 2])
                origin = np.array([x1,y1,zmax])
                lengths = np.array([la, lb, 2.6]) # 2.6 is distance between reactant and surface
                print("origin: ", origin)

                # add reactant
                candidate = self.__place_random_configuration(self.constrained_distance, origin, lengths)
                candidates.append(candidate)
        write("cand.xyz", candidates)

        return candidates

    def __place_random_configuration(self, bond_distance, origin, lengths):
        """ randomly place reactants in a given region
        """
        # place positions
        atoms = self.reactant.copy()
        # adjust bond distance
        vec = atoms[0].position - atoms[1].position
        new_pos = atoms[0].position + vec / np.linalg.norm(vec) * bond_distance
        atoms[1].position = new_pos

        # Apply a random translation
        for i in range(100):
            candidate = self.surface.copy()
            ran_pos = np.random.random(3) 
            ran_pos[2] = 1.0
            pos = ran_pos * lengths + origin
            com = atoms.get_center_of_mass()
            atoms.translate(pos - com) # set to the origin
            # Apply a random rotation to multi-atom blocks
            phi, theta, psi = 360 * np.random.random(3)
            atoms.euler_rotate(
                phi=phi, theta=0.5 * theta, psi=psi,
                center=pos
            )
            # add reactant
            candidate.extend(atoms)
            # TODO: check anchor point (C is under O)
            # check distance
            if not atoms_too_close(candidate, self.blmin):
                # print(f"generate after {i+1} attempts...")
                break
        else:
            candidate = None
        
        return candidate
    
    def __molecular_dynamics(self, atoms):
        """ CMD
        """
        cons = FixAtoms(indices=[a.index for a in atoms if a.position[2] < 4.5])
        atoms.set_constraint(cons)

        MaxwellBoltzmannDistribution(atoms, temperature_K=self.temperature)
        print("initial temperature: ", atoms.get_temperature())
        write("xxx.xyz", atoms)
        dyn = Langevin(
            atoms, 
            timestep = self.timestep*ase.units.fs, 
            temperature_K = self.temperature, 
            friction = 0.002, # TODO: what is the unit?
            fixcm = True
        )
        plumed_worker = AsePlumed(
            atoms, timestep=0,
            in_file = "plumed.dat",
            out_file = "plumed.out"
        )

        md_forces = atoms.get_forces()

        with open("md.xyz", "w") as fopen:
            fopen.write("")

        mdsteps = 500
        for i in range(mdsteps):
            dyn.step(md_forces)

            # bias forces
            energy = atoms.get_potential_energy()
            md_forces = atoms.get_forces()
            bias_forces = plumed_worker.external_forces(
                i, new_energy=energy, new_forces=md_forces
            )
            md_forces = bias_forces

            write("md.xyz", atoms, append=True)
        
        return
    
    def __call__(self, calc):

        # place MD
        if Path("./cand.xyz").exists():
            candidates = read("./cand.xyz", ":")
        else:
            candidates = self.__partition_surface_grid()
        print("number of candidates: ", candidates)

        # run constrained MD
        for cand in candidates[3:]:
            calc.reset()
            cand.calc = calc
            self.__molecular_dynamics(cand)

            # exit()

        return


if __name__ == "__main__":
    reactant = read("/mnt/scratch2/users/40247882/catsign/lasp-main/reactions/CO/allstr.arc", "-1", format="dmol-arc")
    surface = read("/mnt/scratch2/users/40247882/catsign/lasp-main/sample/ZnO-2Ov4Cr/IS/allstr.arc", "-1", format="dmol-arc")

    grs = GridReactionSampling(reactant, surface)

    from gdpx.computation.lasp import LaspNN
    pot_path = "/mnt/scratch2/users/40247882/catsign/lasp-main/ZnCrOCH.pot"
    pot = dict(
        H  = pot_path,
        C  = pot_path,
        O  = pot_path,
        Cr = pot_path,
        Zn = pot_path
    )
    calc = LaspNN(
        directory = "./LaspNN-Worker",
        command = "mpirun -n 4 lasp",
        pot=pot
    )

    grs(calc)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import NoReturn, Callable, List

import numpy as np
import networkx as nx
from joblib import Parallel, delayed

import ase
from ase import Atoms
from ase.io import read, write


from .. import config
from .. import CustomTimer
from .. import SiteFinder
from .modifier import GraphModifier, DEFAULT_GRAPH_PARAMS


def str2atoms(species: str) -> Atoms:
    """Convert a string to an Atoms object."""
    # - build adsorbate
    atoms = None
    if species in ase.data.chemical_symbols:
        atoms = Atoms(species, positions=[[0.0, 0.0, 0.0]])
    elif species in ase.collections.g2.names:
        atoms = ase.build.molecule(species)
    else:
        raise ValueError(f"Fail to create species {species}")

    return atoms


def single_insert_adsorbate(
    graph_params: dict, idx, atoms, ads, site_params: list, 
    print_func: Callable=print, debug_func: Callable=print
):
    """Insert adsorbate into the graph.
    """
    site_creator = SiteFinder(**graph_params)
    site_creator._print = print_func
    site_creator._debug = debug_func
    site_groups = site_creator.find(atoms, site_params)

    ads_indices = [a.index for a in atoms if a.symbol in site_creator.adsorbate_elements]

    created_frames = []
    for i, (sites, params) in enumerate(zip(site_groups, site_params)):
        ads_params = params.get("ads", [{}])
        cur_frames = []
        for s in sites: 
            ads_frames = s.adsorb(
                ads, ads_indices, ads_params
            )
            cur_frames.extend(ads_frames)
        created_frames.extend(cur_frames)
        print_func(
            f"group {i} unique sites {len(sites)} with {len(cur_frames)} frames for substrate {idx}."
        )
    
    return created_frames


class GraphInsertModifier(GraphModifier):

    name = "graph_insert"

    def __init__(
            self, species, spectators: List[str], 
            sites: List[dict], substrates=None, graph: dict = DEFAULT_GRAPH_PARAMS,
            *args, **kwargs
        ):
        """Insert an adsorbate on sites according to graph representation.
        """
        super().__init__(substrates=substrates, *args, **kwargs)

        self.species = species # make this a node

        self.spectators = spectators 
        self.site_params = sites
        self.graph_params = graph

        #self.check_site_unique = True
        # adsorbate_indices
        # site_radius
        # region

        # - parallel
        self.njobs = config.NJOBS

        return
    
    def _irun(self, substrates: List[Atoms]) -> List[Atoms]:
        """Insert an adsorabte on the substrate."""
        # -- parameters for finding adsorption sites
        site_params = copy.deepcopy(self.site_params)

        # -- parameters for site graph
        graph_params = copy.deepcopy(self.graph_params) # to create site graph
        adsorbate_elements = copy.deepcopy(self.spectators)
        graph_params.update( 
            dict(
                adsorbate_elements = adsorbate_elements,
                #coordination_numbers = params.get("coordination_numbers"),
                site_radius = 2,
                #check_site_unique = params.get("check_site_unique", True)
            )
        )

        # -- parameters for species used for comparison
        species = self.species # to insert
        self._print(f"start to insert adsorbate {species}.")

        # - build adsorbate (single atom or molecule)
        #   and update selected_species
        if isinstance(species, str):
            # simple species
            adsorbate = str2atoms(species)
        else: # dict
            adsorbate = read(species["adsorbate"]) # only one structure
        symbols = list(set(adsorbate.get_chemical_symbols()))

        selected_species = copy.deepcopy(graph_params.get("adsorbate_elements", []))
        selected_species.extend(symbols)
        selected_species = list(set(selected_species))

        # - get structures with inserted species
        with CustomTimer(name="insert-adsorbate", func=self._print):
            #ret = Parallel(n_jobs=self.njobs)(
            ret = Parallel(n_jobs=1)(
                delayed(single_insert_adsorbate)(
                    graph_params, idx, a, adsorbate, site_params,
                    print_func=self._print, debug_func=self._debug
                ) for idx, a in enumerate(substrates)
            )
        ret_frames = []
        for frames in ret:
            ret_frames.extend(frames)
        write(self.directory/f"possible_frames.xyz", ret_frames)
        self._print(f"nframes of inserted: {len(ret_frames)}")

        # NOTE: It is unnecessary to compare among substrates if the spectator
        #       adsorbates are not the same as the inserted one. Otherwise, 
        #       comparasion should be performed.
        target_group = ["symbol "+" ".join(selected_species)]
        created_frames = self._compare_structures(
            ret_frames, graph_params, target_group
        )

        return created_frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ase.ga.ofp_comparator import OFPComparator
from ase.ga.particle_comparator import NNMatComparator
from ase.ga.standard_comparators import InteratomicDistanceComparator

from ..core.register import registers

registers.comparator.register(OFPComparator)
registers.comparator.register(NNMatComparator)
registers.comparator.register(InteratomicDistanceComparator)

from .cartesian import CartesianComparator
registers.comparator.register(CartesianComparator)

from .coordination import CoordinationComparator
registers.comparator.register(CoordinationComparator)

from .graph import GraphComparator
registers.comparator.register(GraphComparator)

from .singlepoint import SinglePointComparator
registers.comparator.register("single_point")(SinglePointComparator)

from .reaction import ReactionComparator
registers.comparator.register("reaction")(ReactionComparator)



if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import warnings

from .. import config
from ..utils.command import CustomTimer

from .interface import create_builder

from ..core.register import registers
from .direct import DirectBuilder, ReadBuilder
from .species import MoleculeBuilder
from .dimer import DimerBuilder


# - regions
from .region import (
    AutoRegion, CubeRegion, SphereRegion, CylinderRegion, LatticeRegion,
    SurfaceLatticeRegion, SurfaceRegion,
    IntersectRegion
)
registers.region.register(AutoRegion)
registers.region.register(CubeRegion)
registers.region.register(SphereRegion)
registers.region.register(CylinderRegion)
registers.region.register(LatticeRegion)
registers.region.register(SurfaceLatticeRegion) # BUG?
registers.region.register(SurfaceRegion)        # BUG?
registers.region.register(IntersectRegion)


# - basic builders and modifiers
registers.builder.register("direct")(DirectBuilder)
registers.builder.register("reader")(ReadBuilder)
registers.builder.register("dimer")(DimerBuilder)
registers.builder.register("molecule")(MoleculeBuilder)

from .perturbator import PerturbatorBuilder
registers.builder.register("perturb")(PerturbatorBuilder)

from .packer import PackerBuilder
registers.builder.register("pack")(PackerBuilder)

from .insert import InsertModifier
registers.builder.register("insert")(InsertModifier)

# --
from ..graph.sites import SiteFinder
from ..graph.creator import StruGraphCreator, extract_chem_envs
from ..graph.comparison import get_unique_environments_based_on_bonds, paragroup_unique_chem_envs
from ..graph.utils import unpack_node_name

from .graph import GraphInsertModifier, GraphRemoveModifier, GraphExchangeModifier
registers.builder.register("graph_insert")(GraphInsertModifier)
registers.builder.register("graph_remove")(GraphRemoveModifier)
registers.builder.register("graph_exchange")(GraphExchangeModifier)

# --
from .randomBuilder import BulkBuilder, ClusterBuilder, SurfaceBuilder
registers.builder.register("random_bulk")(BulkBuilder)
registers.builder.register("random_cluster")(ClusterBuilder)
registers.builder.register("random_surface")(SurfaceBuilder)

from .cleave_surface import CleaveSurfaceModifier
registers.builder.register("cleave_surface")(CleaveSurfaceModifier)

from .repeat import RepeatModifier
registers.builder.register("repeat")(RepeatModifier)

# - extra modifiers
from .zoom import ZoomModifier
registers.builder.register("zoom")(ZoomModifier)


# - optional
try:
    from .hypercube import HypercubeBuilder
    registers.builder.register("hypercube")(HypercubeBuilder)
except ImportError as e:
    warnings.warn("Module {} import failed: {}".format("hypercube", e), UserWarning)

# - extra utilities
from .utils import remove_vacuum, reset_cell
registers.operation.register(remove_vacuum)
registers.operation.register(reset_cell)


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import numpy as np
import os
import torch
import re

# from gpu_sel import *
from ase.data import chemical_symbols, atomic_numbers
from ase.units import Bohr
from ase.calculators.calculator import (
    Calculator,
    all_changes,
    PropertyNotImplementedError,
)

"""Reann Calculator.

This is the same as the official one. However, we use a lazy initialisation of 
model here, which avoids serialisation problem when copying this calculator.

"""


class REANN(Calculator):

    implemented_properties = ["energy", "forces"]

    def __init__(
        self,
        atomtype,
        maxneigh,
        getneigh,
        nn="PES.pt",
        device="cpu",
        dtype=torch.float32,
        **kwargs,
    ):
        Calculator.__init__(self, **kwargs)
        self.device = torch.device(device)
        self.dtype = dtype
        self.atomtype = atomtype
        self.maxneigh = maxneigh
        self.getneigh = getneigh

        # - lazy init
        self._nn_path = nn
        self.pes = None
        self.cutoff = None

        self.table = 0
        # self.pes=torch.compile(pes)

    def _init_model(self) -> None:
        """Lazy import some attributes as they not picklable."""
        pes = torch.jit.load(self._nn_path)
        pes.to(self.device).to(self.dtype)
        pes.eval()
        self.cutoff = pes.cutoff
        self.pes = torch.jit.optimize_for_inference(pes)

        from reann.ASE import getneigh
        self.getneigh = getneigh

        return

    def calculate(
        self, atoms=None, properties=["energy", "force"], system_changes=all_changes
    ):
        Calculator.calculate(self, atoms, properties, system_changes)
        if self.pes is None:
            self._init_model()

        # TODO: To re-use the same calculator, we init neigh_list every time
        cell = np.array(self.atoms.cell)
        #if self.table > 0.5 and "cell" in system_changes:
        #    self.getneigh.deallocate_all()
        #if "cell" in system_changes:
        #    if cell.ndim == 1:
        #        cell = np.diag(cell)
        #    self.getneigh.init_neigh(self.cutoff, self.cutoff / 2.0, cell.T)
        #    self.table += 1
        if self.getneigh.initmod.rc > 0.:
            self.getneigh.deallocate_all()

        if cell.ndim == 1:
            cell = np.diag(cell)
        self.getneigh.init_neigh(self.cutoff, self.cutoff / 2.0, cell.T)
        self.table += 1

        icart = self.atoms.get_positions()
        cart, neighlist, shiftimage, scutnum = self.getneigh.get_neigh(
            icart.T, self.maxneigh
        )
        cart = torch.from_numpy(cart.T).contiguous().to(self.device).to(self.dtype)
        neighlist = (
            torch.from_numpy(neighlist[:, :scutnum])
            .contiguous()
            .to(self.device)
            .to(torch.long)
        )
        shifts = (
            torch.from_numpy(shiftimage.T[:scutnum, :])
            .contiguous()
            .to(self.device)
            .to(self.dtype)
        )
        symbols = list(self.atoms.symbols)
        species = [self.atomtype.index(i) for i in symbols]
        species = torch.tensor(species, device=self.device, dtype=torch.long)
        energy, force = self.pes(cart, neighlist, shifts, species)
        energy = float(energy.detach().numpy())
        self.results["energy"] = energy
        force = force.detach().numpy()
        self.results["forces"] = force.copy()


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List

import numpy as np

from ase import Atoms
from ase.io import read, write

from .builder import StructureModifier


class ZoomModifier(StructureModifier):

    """Extend or compress bulk.
    """

    def __init__(self, coefs: List[float]=None, substrates=None, *args, **kwargs):
        """"""
        super().__init__(substrates=substrates, *args, **kwargs)
        if coefs is None:
            coefs = np.arange(0.6, 1.8, 0.05)
        self.coefs = coefs

        return
    
    def run(self, substrates=None, size: int=1, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        frames = []
        for substrate in self.substrates:
            curr_frames = self._irun(substrate=substrate, size=size, *args, **kwargs)
            frames.extend(curr_frames)

        return frames
    
    def _irun(self, substrate: Atoms, size: int, *args, **kwargs) -> List[Atoms]:
        """"""
        volume = substrate.get_volume()
        cell = copy.deepcopy(substrate.get_cell(complete=True))

        frames = []
        for i in self.coefs:
            atoms = copy.deepcopy(substrate)
            atoms.set_cell(cell*(i)**(1/3.), scale_atoms=True)
            frames.append(atoms)

        return frames
    

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
import pathlib
from typing import NoReturn, Optional, List, Callable

import omegaconf

from ase.calculators.calculator import Calculator

from ..core.operation import Operation
from ..core.variable import Variable
from ..core.register import registers
from ..utils.command import parse_input_file

from ..potential.manager import AbstractPotentialManager
from ..scheduler.scheduler import AbstractScheduler
from .worker import AbstractWorker
from .drive import DriverBasedWorker, CommandDriverBasedWorker, QueueDriverBasedWorker
from .react import ReactorBasedWorker
from .single import SingleWorker


def convert_input_to_potter(
    inp,
    estimate_uncertainty: Optional[bool] = False,
    switch_backend: Optional[str] = None,
    print_func: Callable = print,
) -> List[AbstractPotentialManager]:
    """Convert an input to a potter and adjust its behaviour."""
    potter = None
    if isinstance(inp, AbstractPotentialManager):
        potter = inp
    elif isinstance(inp, Variable):
        potter = inp.value
    elif isinstance(inp, dict) or isinstance(inp, omegaconf.dictconfig.DictConfig):
        potter_params = copy.deepcopy(inp)
        name = potter_params.get("name", None)
        potter = registers.create(
            "manager",
            name,
            convert_name=True,
        )
        potter.register_calculator(potter_params.get("params", {}))
        potter.version = potter_params.get("version", "unknown")
    elif isinstance(inp, str) or isinstance(inp, pathlib.Path):
        if pathlib.Path(inp).exists():
            potter_params = parse_input_file(input_fpath=inp)
            name = potter_params.get("name", None)
            potter = registers.create(
                "manager",
                name,
                convert_name=True,
            )
            potter.register_calculator(potter_params.get("params", {}))
            potter.version = potter_params.get("version", "unknown")
        else:
            raise RuntimeError(f"The potter configuration `{inp}` does not exist.")
    else:
        raise RuntimeError(f"Unknown {inp} of type {type(inp)} for the potter.")

    # HACK: broadcast potters

    # if isinstance(potter.calc, list):
    #     num_calculators = len(potter.calc)
    #     assert num_calculators > 1
    #     calcs = potter.calc
    #     potter.calc = None
    #     potters = []
    #     for i in range(num_calculators):
    #         p = copy.deepcopy(potter)
    #         p.calc = calcs[i]
    #         potters.append(p)
    # else:
    #     num_calculators = 1
    #     potters = [potter]

    if hasattr(potter, "broadcast"):
        potters = potter.broadcast(potter)
        # print_func(f"{potters =}")
    else:
        potters = [potter]

    for p in potters:
        assert isinstance(p.calc, Calculator)

    # adjust potter behaviour
    for i, potter in enumerate(potters):
        print_func(f"potter-{i} {potter.name}")
        if hasattr(potter, "switch_uncertainty_estimation"):
            if estimate_uncertainty is not None:
                print_func(
                    f"{potter.name} switches its uncertainty estimation to {estimate_uncertainty}..."
                )
                potter.switch_uncertainty_estimation(estimate_uncertainty)
            else:
                ...
        else:
            print_func(
                f"{potter.name} does not support switching its uncertainty estimation..."
            )

        if hasattr(potter, "switch_backend"):
            if switch_backend is not None:
                print_func(f"{potter.name} switches its backend to {switch_backend}...")
                potter.switch_backend(backend=switch_backend)
            else:
                ...
        else:
            print_func(f"{potter.name} does not support switching its backend...")

    return potters


@registers.variable.register
class ComputerVariable(Variable):

    def __init__(
        self,
        potter,
        driver={},
        scheduler={},
        *,
        estimate_uncertainty: Optional[bool] = None,
        switch_backend: Optional[str] = None,
        batchsize: int = 1,
        share_wdir: bool = False,
        use_single: bool = False,
        retain_info: bool = False,
        directory=pathlib.Path.cwd(),
    ):
        """"""
        self.potter = convert_input_to_potter(
            potter,
            estimate_uncertainty=estimate_uncertainty,
            switch_backend=switch_backend,
            print_func=self._print,
        )
        self.driver = self._load_driver(driver)
        self.scheduler = self._load_scheduler(scheduler)

        # - ...
        self.batchsize = batchsize  # NOTE: This can be updated in drive operation.

        workers = self._broadcast_workers(
            self.potter,
            self.driver,
            self.scheduler,
            batchsize=self.batchsize,
            share_wdir=share_wdir,
            use_single=use_single,
            retain_info=retain_info,
        )
        super().__init__(workers)

        self.use_single = use_single

        return

    def _load_driver(self, inp) -> List[dict]:
        """Load drivers from a Variable or a dict."""
        # print("driver: ", inp)
        drivers = []  # params
        if isinstance(inp, Variable):
            drivers = inp.value
        elif isinstance(inp, list):  # assume it contains a List of dicts
            drivers = inp
        elif isinstance(inp, dict) or isinstance(
            inp, omegaconf.dictconfig.DictConfig
        ):  # assume it only contains one driver
            driver_params = copy.deepcopy(inp)
            # driver = self.potter.create_driver(driver_params) # use external backend
            drivers = [driver_params]
        else:
            raise RuntimeError(f"Unknown {inp} for drivers.")

        return drivers

    def _load_scheduler(self, inp):
        """"""
        scheduler = None
        if isinstance(inp, AbstractScheduler):
            scheduler = inp
        elif isinstance(inp, Variable):
            scheduler = inp.value
        elif isinstance(inp, dict):
            scheduler_params = copy.deepcopy(inp)
            backend = scheduler_params.pop("backend", "local")
            scheduler = registers.create(
                "scheduler", backend, convert_name=True, **scheduler_params
            )
        else:
            raise RuntimeError(f"Unknown {inp} for the scheduler.")

        return scheduler

    def _broadcast_workers(
        self,
        potters,
        drivers,
        scheduler,
        *,
        batchsize: int = 1,
        share_wdir: bool = False,
        use_single: bool = False,
        retain_info: bool = False,
    ) -> List[AbstractWorker]:
        """Create a list of workers."""
        # check potters
        num_potters = len(potters)
        self._print(f"{num_potters =}")

        # check if there were custom wdirs, and zip longest
        num_drivers = len(drivers)

        # broadcast
        pairs = list(itertools.product(range(num_potters), range(num_drivers)))
        num_pairs = len(pairs)

        wdirs = [self.directory / f"w{i}" for i in range(num_pairs)]

        # create workers
        workers = []
        for i, (p_i, d_i) in enumerate(pairs):
            # workers share calculator in potter
            driver = potters[p_i].create_driver(drivers[d_i])
            if not use_single:
                if scheduler.name == "local":
                    worker = CommandDriverBasedWorker(potters[p_i], driver, scheduler)
                else:
                    worker = QueueDriverBasedWorker(potters[p_i], driver, scheduler)
            else:
                worker = SingleWorker(potters[d_i], driver, scheduler)
            worker._share_wdir = share_wdir
            worker._retain_info = retain_info
            # wdir is temporary as it may be reset by drive operation
            worker.directory = wdirs[i]
            workers.append(worker)

        for worker in workers:
            worker.batchsize = batchsize

        return workers


@registers.variable.register
class ReactorVariable(Variable):
    """Create a ReactorBasedWorker.

    TODO:
        Broadcast driver params to give several workers?

    """

    def __init__(
        self,
        potter,
        driver: dict,
        scheduler={},
        *,
        estimate_uncertainty: Optional[bool] = None,
        switch_backend: Optional[str] = None,
        batchsize=1,
        directory="./",
    ):
        """"""
        # - save state by all nodes
        self.potter = convert_input_to_potter(
            potter,
            estimate_uncertainty=estimate_uncertainty,
            switch_backend=switch_backend,
            print_func=self._print,
        )
        self.driver = self._load_driver(driver)
        self.scheduler = self._load_scheduler(scheduler)

        self.batchsize = batchsize

        # - create a reactor
        # reactor = self.potter.create_reactor(kwargs)
        workers = self._create_workers(
            self.potter, self.driver, self.scheduler, batchsize=self.batchsize
        )

        super().__init__(initial_value=workers, directory=directory)

        return

    def _load_driver(self, inp) -> List[dict]:
        """Load drivers from a Variable or a dict."""
        # print("driver: ", inp)
        drivers = []  # params
        if isinstance(inp, Variable):
            drivers = inp.value
        elif isinstance(inp, list):  # assume it contains a List of dicts
            drivers = inp
        elif isinstance(inp, dict) or isinstance(inp, omegaconf.dictconfig.DictConfig):
            driver_params = copy.deepcopy(inp)
            # driver = self.potter.create_driver(driver_params) # use external backend
            drivers = [driver_params]
        else:
            raise RuntimeError(f"Unknown {inp} for drivers.")

        return drivers

    def _load_scheduler(self, inp):
        """"""
        scheduler = None
        if isinstance(inp, Variable):
            scheduler = inp.value
        elif isinstance(inp, dict):
            scheduler_params = copy.deepcopy(inp)
            backend = scheduler_params.pop("backend", "local")
            scheduler = registers.create(
                "scheduler", backend, convert_name=True, **scheduler_params
            )
        else:
            raise RuntimeError(f"Unknown {inp} for the scheduler.")

        return scheduler

    def _create_workers(
        self,
        potters,
        drivers: List[dict],
        scheduler,
        *,
        batchsize: int = 1,
    ):
        """"""
        # FIXME: Support several potters?
        num_potters = len(potters)
        assert num_potters == 1, "Reactor only supports one potter."
        potter = potters[0]

        workers = []
        for driver_params in drivers:
            driver = potter.create_reactor(driver_params)
            worker = ReactorBasedWorker(potter, driver, scheduler)
            workers.append(worker)

        for worker in workers:
            worker.batchsize = batchsize

        return workers


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


import copy

from . import AbstractPotentialManager


class AsePotManager(AbstractPotentialManager):

    name = "ase"

    implemented_backends = ["ase"]
    valid_combinations = (
        ("ase", "ase"),
    )

    """Here is an interface to ase built-in calculators."""

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return

    def register_calculator(self, calc_params: dict, *args, **kwargs):
        """"""
        super().register_calculator(calc_params, *args, **kwargs)

        calc_params = copy.deepcopy(calc_params)
        method = calc_params.pop("method", "")

        if self.calc_backend == "ase":
            if method == "lj":
                from ase.calculators.lj import LennardJones
                calc_cls = LennardJones
            elif method == "morse":
                from ase.calculators.morse import MorsePotential
                calc_cls = MorsePotential
            elif method == "tip3p":
                from ase.calculators.tip3p import TIP3P
                calc_cls = TIP3P
            else:
                raise NotImplementedError(f"Unsupported potential {method}.")
        else:
            raise NotImplementedError(f"Unsupported backend {self.calc_backend}.")
        
        calc = calc_cls(**calc_params)

        self.calc = calc

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import pathlib
import time
import uuid
import yaml

from typing import List

from tinydb import Query, TinyDB

from ase import Atoms
from ase.io import read, write

from .worker import AbstractWorker
from ..computation.driver import AbstractDriver
from ..potential.manager import AbstractPotentialManager
from ..utils.command import CustomTimer
from ..scheduler.scheduler import AbstractScheduler
from ..scheduler.local import LocalScheduler


class GridDriverBasedWorker(AbstractWorker):

    def __init__(self, potter: List[AbstractPotentialManager], driver: AbstractDriver, scheduler: AbstractScheduler=LocalScheduler(), directory="./", *args, **kwargs) -> None:
        """"""
        super().__init__(directory)

        self.potter = potter # a List of potters
        self.driver = driver
        self.scheduler = scheduler

        return
    
    def run(self, builder, *args, **kwargs):
        """This worker accepts only a single structure."""
        super().run(*args, **kwargs)

        # - structures
        if isinstance(builder, list): # assume List[Atoms]
            frames = builder
        else: # assume it is a builder
            frames = builder.run()
        nframes = len(frames)
        assert len(frames) == 1, f"{self.__class__.__name__} accepts only a single structure."

        # - potentials
        driver_dict = self.driver.as_dict()
        drivers = [p.create_driver(driver_dict) for p in self.potter]
        ndrivers = len(drivers)

        # -

        uid = str(uuid.uuid1())
        wdir_names = [f"pott{i}" for i in range(ndrivers)]
        job_name = uid

        scheduler = self.scheduler
        if scheduler.name == "local":
            with CustomTimer(name="run-driver", func=self._print):
                for wdir_name, driver in zip(wdir_names, drivers):
                    driver.directory = self.directory / wdir_name
                    self._print(
                        f"{time.asctime( time.localtime(time.time()) )} {wdir_name} {driver.directory.name} is running..."
                    )
                    driver.reset()
                    driver.run(frames[0], read_ckpt=True, extra_info=None)
        else:
            worker_params = {}
            worker_params["use_single"] = True
            worker_params["driver"] = self.driver.as_dict()
            worker_params["potential"] = self.potter.as_dict()

            with open(self.directory/f"worker-{uid}.yaml", "w") as fopen:
                yaml.dump(worker_params, fopen)

            # - save structures
            dataset_path = str((self.directory/f"_gdp_inp.xyz").resolve())
            write(dataset_path, frames[0])

            # - save scheduler file
            jobscript_fname = f"run-{uid}.script"
            self.scheduler.job_name = job_name
            self.scheduler.script = self.directory/jobscript_fname

            self.scheduler.user_commands = "gdp -p {} compute {}\n".format(
                (self.directory/f"worker-{uid}.yaml").name, dataset_path
            )

            # - TODO: check whether params for scheduler is changed
            self.scheduler.write()
            if self._submit:
                self._print(f"{self.directory.name} JOBID: {self.scheduler.submit()}")
            else:
                self._print(f"{self.directory.name} waits to submit.")

        # - save this batch job to the database
        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            _ = database.insert(
                dict(
                    uid = uid,
                    #md5 = identifier,
                    gdir=job_name, 
                    #group_number=ig, 
                    wdir_names=wdir_names,
                    queued=True
                )
            )

        return

    def inspect(self, resubmit=False, *args, **kwargs):
        """"""
        self._initialise(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs()

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in running_jobs:
                doc_data = database.get(Query().gdir == job_name)
                uid = doc_data["uid"]

                self.scheduler.job_name = job_name
                self.scheduler.script = self.directory/"train.script"

                if self.scheduler.is_finished():
                    # -- check if the job finished properly
                    is_finished = False
                    wdir_names = doc_data["wdir_names"]
                    for x in wdir_names:
                        wdir = self.directory/x
                        if not wdir.exists():
                            break
                        else:
                            self.driver.directory = wdir
                            if not self.driver.read_convergence():
                                break
                    else:
                        is_finished = True
                    if is_finished:
                        self._print(f"{job_name} is finished.")
                        database.update({"finished": True}, doc_ids=[doc_data.doc_id])
                    else:
                        if resubmit:
                            jobid = self.scheduler.submit()
                            self._print(f"{job_name} is re-submitted with JOBID {jobid}.")
                else:
                    self._print(f"{job_name} is running...")

        return

    def retrieve(self, include_retrieved: bool=False, *args, **kwargs):
        """Retrieve training results.

        """
        self.inspect(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+retrieve")

        unretrieved_wdirs_ = []
        if not include_retrieved:
            unretrieved_jobs = self._get_unretrieved_jobs()
        else:
            unretrieved_jobs = self._get_finished_jobs()

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                unretrieved_wdirs_.extend(
                    (self.directory/w).resolve() for w in doc_data["wdir_names"]
                )
            unretrieved_wdirs = unretrieved_wdirs_

            results = []
            if unretrieved_wdirs:
                unretrieved_wdirs = [pathlib.Path(x) for x in unretrieved_wdirs]
                self._debug(f"unretrieved_wdirs: {unretrieved_wdirs}")
                for p in unretrieved_wdirs:
                    self.driver.directory = p
                    results.append(self.driver.read_trajectory())

            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                database.update({"retrieved": True}, doc_ids=[doc_data.doc_id])

        return results
    
    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()

        return params
    

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import time
from typing import Optional, List, Mapping

import omegaconf

from ase import Atoms

from ..core.register import registers
from ..core.variable import Variable
from ..core.operation import Operation
from ..data.array import AtomsNDArray


@registers.operation.register
class pair_stru(Operation):

    def __init__(self, structures, method="couple", pairs=None, directory="./") -> None:
        """"""
        super().__init__(input_nodes=[structures], directory=directory)

        assert method in [
            "couple",
            "concat",
            "custom",
        ], "pair method should be either couple or concat."
        self.method = method
        self.pairs = pairs

        return

    def forward(self, structures: AtomsNDArray) -> List[List[Atoms]]:
        """"""
        super().forward()

        # NOTE: assume this is a 1-D array
        self._debug(f"structures: {structures}")
        intermediates = structures.get_marked_structures()

        nframes = len(intermediates)
        rankings = list(range(nframes))

        if self.method == "couple":
            pair_indices = [rankings[0::2], rankings[1::2]]
        elif self.method == "concat":
            pair_indices = [rankings[:-1], rankings[1:]]
        elif self.method == "custom":
            pair_indices = self.pairs
        else:
            ...

        pair_structures = []
        for p in pair_indices:
            pair_structures.append([intermediates[i] for i in p])

        # Must update status at the end of forward! Otherwise, the status will be 
        # overwritten in active session by node.reset().
        self.status = "finished"

        return AtomsNDArray(pair_structures)


@registers.operation.register
class react(Operation):

    def __init__(
        self, structures, reactor, batchsize: Optional[int] = None, directory="./"
    ) -> None:
        """"""
        super().__init__(input_nodes=[structures, reactor], directory=directory)

        self.batchsize = batchsize

        return

    def forward(self, structures: AtomsNDArray, reactors):
        """"""
        super().forward()

        # - assume structures be an AtomsNDArray with a shape of (num_pairs, 2)
        # structures = [[x[-1] for x in s] for s in structures]
        # if isinstance(structures, list):
        #     # - from list_nodes operation
        #     structures = AtomsNDArray([x.tolist() for x in structures])
        # else:
        #     # - from pair operation
        #     structures = AtomsNDArray(structures)
        nreactions = len(structures)

        # - create reactors
        for i, reactor in enumerate(reactors):
            reactor.directory = self.directory / f"r{i}"
            if self.batchsize is not None:
                reactor.batchsize = self.batchsize
            else:
                reactor.batchsize = nreactions
        nreactors = len(reactors)

        if nreactors == 1:
            reactors[0].directory = self.directory

        # - align structures
        reactor_status = []
        for i, reactor in enumerate(reactors):
            flag_fpath = reactor.directory / "FINISHED"
            self._print(f"run reactor {i} for {nreactions} nframes")
            if not flag_fpath.exists():
                reactor.run(structures)
                reactor.inspect(resubmit=True)
                if reactor.get_number_of_running_jobs() == 0:
                    with open(flag_fpath, "w") as fopen:
                        fopen.write(
                            f"FINISHED AT {time.asctime( time.localtime(time.time()) )}."
                        )
                    reactor_status.append(True)
                else:
                    reactor_status.append(False)
            else:
                with open(flag_fpath, "r") as fopen:
                    content = fopen.readlines()
                self._print(content)
                reactor_status.append(True)

        if all(reactor_status):
            self.status = "finished"

        return reactors


if __name__ == "__main__":
    ...
